#!/usr/bin/env python
# Copyright 2020-2021 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Authors: Qiming Sun <osirpt.sun@gmail.com>
#

'''
Range separation JK builder

Ref:
    Q. Sun, arXiv:2012.07929
    Q. Sun, arXiv:2302.11307
'''

import ctypes
import numpy as np
import scipy.linalg
from pyscf import gto
from pyscf import lib
from pyscf.lib import logger
from pyscf.scf import _vhf
from pyscf.pbc import gto as pbcgto
from pyscf.pbc.tools import pbc as pbctools
from pyscf.pbc.tools import k2gamma
from pyscf.pbc.scf import addons
from pyscf.pbc.df import aft, rsdf_builder, aft_jk
from pyscf.pbc.df import ft_ao
from pyscf.pbc.df.df_jk import (zdotNN, zdotCN, zdotNC, _ewald_exxdiv_for_G0,
                                _format_dms, _format_kpts_band, _format_jks)
from pyscf.pbc.df.incore import libpbc, _get_cache_size, LOG_ADJUST
from pyscf.pbc.lib.kpts_helper import (is_zero, kk_adapted_iter,
                                       group_by_conj_pairs, intersection)
from pyscf import __config__

# Threshold of steep bases and local bases
RCUT_THRESHOLD = getattr(__config__, 'pbc_scf_rsjk_rcut_threshold', 1.0)
OMEGA_MIN = rsdf_builder.OMEGA_MIN
INDEX_MIN = rsdf_builder.INDEX_MIN

class RangeSeparatedJKBuilder(lib.StreamObject):
    _keys = {
        'cell', 'mesh', 'kpts', 'purify', 'omega', 'rs_cell', 'cell_d',
        'bvk_kmesh', 'supmol_sr', 'supmol_ft', 'supmol_d', 'cell0_basis_mask',
        'ke_cutoff', 'direct_scf_tol', 'time_reversal_symmetry',
        'exclude_dd_block', 'allow_drv_nodddd', 'approx_vk_lr_missing_mo',
    }

    def __init__(self, cell, kpts=np.zeros((1,3))):
        self.cell = cell
        self.stdout = cell.stdout
        self.verbose = cell.verbose
        self.max_memory = cell.max_memory
        self.mesh = None
        self.kpts = np.reshape(kpts, (-1, 3))
        self.purify = True

        if cell.omega == 0:
            self.omega = None
        elif cell.omega < 0:
            # Initialize omega to cell.omega for HF exchange of short range
            # int2e in RSH functionals
            self.omega = -cell.omega
        else:
            raise RuntimeError('RSJK does not support LR integrals')
        self.rs_cell = None
        self.cell_d = None
        # Born-von Karman supercell
        self.bvk_kmesh = None
        self.supmol_sr = None
        self.supmol_ft = None
        self.supmol_d = None
        # which shells are located in the first primitive cell
        self.cell0_basis_mask = None
        self.ke_cutoff = None
        self.direct_scf_tol = None
        # Use the k-point conjugation symmetry between k and -k
        self.time_reversal_symmetry = True
        self.exclude_dd_block = True
        self.allow_drv_nodddd = True
        self._sr_without_dddd = None
        # Allow to approximate vk_lr with less mesh if mo_coeff not attached to dm.
        self.approx_vk_lr_missing_mo = False
        # TODO: incrementally build SR part
        self._last_vs = (0, 0)
        self._qindex = None

    __getstate__, __setstate__ = lib.generate_pickle_methods(
            excludes=('rs_cell', 'cell_d', 'supmol_sr', 'supmol_ft', 'supmol_d',
                      '_sr_without_dddd', '_last_vs', '_qindex'),
            reset_state=True)

    def has_long_range(self):
        '''Whether to add the long-range part computed with AFT/FFT integrals'''
        return self.omega is None or abs(self.cell.omega) < self.omega

    def dump_flags(self, verbose=None):
        logger.info(self, '\n')
        logger.info(self, '******** %s ********', self.__class__)
        logger.info(self, 'mesh = %s (%d PWs)', self.mesh, np.prod(self.mesh))
        logger.info(self, 'omega = %s', self.omega)
        logger.info(self, 'purify = %s', self.purify)
        logger.info(self, 'bvk_kmesh = %s', self.bvk_kmesh)
        logger.info(self, 'ke_cutoff = %s', self.ke_cutoff)
        logger.info(self, 'time_reversal_symmetry = %s', self.time_reversal_symmetry)
        logger.info(self, 'has_long_range = %s', self.has_long_range())
        logger.info(self, 'exclude_dd_block = %s', self.exclude_dd_block)
        logger.info(self, 'allow_drv_nodddd = %s', self.allow_drv_nodddd)
        logger.debug(self, 'approx_vk_lr_missing_mo = %s', self.approx_vk_lr_missing_mo)
        return self

    def reset(self, cell=None):
        if cell is not None:
            self.cell = cell
        self.rs_cell = None
        self.cell_d = None
        self.supmol_sr = None
        self.supmol_ft = None
        self.supmol_d = None
        self._sr_without_dddd = None
        self._last_vs = (0, 0)
        self._qindex = None
        return self

    def build(self, omega=None, intor='int2e'):
        cpu0 = logger.process_clock(), logger.perf_counter()
        log = logger.new_logger(self)
        cell = self.cell
        kpts = self.kpts

        if omega is not None:
            self.omega = omega

        if self.omega is None or self.omega == 0:
            # Search a proper range-separation parameter omega that can balance the
            # computational cost between the real space integrals and moment space
            # integrals
            self.omega, self.mesh, self.ke_cutoff = _guess_omega(cell, kpts, self.mesh)
        else:
            self.ke_cutoff = estimate_ke_cutoff_for_omega(cell, self.omega)
            self.mesh = cell.cutoff_to_mesh(self.ke_cutoff)

        if cell.dimension == 2 and cell.low_dim_ft_type != 'inf_vacuum':
            # To ensure trunc-coulG converged for all basis
            self.mesh[2] = rsdf_builder._estimate_meshz(cell)

        self.check_sanity()

        log.info('omega = %.15g  ke_cutoff = %s  mesh = %s',
                 self.omega, self.ke_cutoff, self.mesh)

        # AFT part exchange cost may be dominant
        if cell.nao * len(kpts) > 5000:
            log.debug1('Disable exclude_dd_block')
            self.exclude_dd_block = False

        # basis with cutoff under ~150 eV are handled by AFTDF
        ke_cutoff = max(6., self.ke_cutoff)
        rs_cell = ft_ao._RangeSeparatedCell.from_cell(
            cell, ke_cutoff, RCUT_THRESHOLD, in_rsjk=True, verbose=log)
        self.rs_cell = rs_cell

        self.bvk_kmesh = kmesh = k2gamma.kpts_to_kmesh(cell, kpts)
        log.debug('kmesh for bvk-cell = %s', kmesh)

        exp_min = np.hstack(cell.bas_exps()).min()
        lattice_sum_factor = max((2*cell.rcut)**3/cell.vol * 1/exp_min, 1)
        cutoff = cell.precision / lattice_sum_factor * .1
        self.direct_scf_tol = cutoff
        log.debug('Set RangeSeparationJKBuilder.direct_scf_tol to %g', cutoff)

        rcut_sr = estimate_rcut(rs_cell, self.omega,
                                exclude_dd_block=self.exclude_dd_block)
        supmol_sr = ft_ao.ExtendedMole.from_cell(rs_cell, kmesh, rcut_sr.max(), log)
        supmol_sr.omega = -self.omega
        self.supmol_sr = supmol_sr.strip_basis(rcut_sr)
        log.debug('supmol nbas = %d cGTO = %d pGTO = %d',
                  supmol_sr.nbas, supmol_sr.nao, supmol_sr.npgto_nr())

        if self.has_long_range():
            rcut = rsdf_builder.estimate_ft_rcut(
                rs_cell, exclude_dd_block=self.exclude_dd_block)
            supmol_ft = rsdf_builder._ExtendedMoleFT.from_cell(
                rs_cell, kmesh, rcut.max(), log)
            supmol_ft.exclude_dd_block = self.exclude_dd_block
            self.supmol_ft = supmol_ft.strip_basis(rcut)
            log.debug('sup-mol-ft nbas = %d cGTO = %d pGTO = %d',
                      supmol_ft.nbas, supmol_ft.nao, supmol_ft.npgto_nr())

            self.cell_d = rs_cell.smooth_basis_cell()
            if self.cell_d.nbas > 0:
                rcut = ft_ao.estimate_rcut(self.cell_d)
                supmol_d = ft_ao.ExtendedMole.from_cell(
                    self.cell_d, self.bvk_kmesh, rcut.max(), log)
                self.supmol_d = supmol_d.strip_basis(rcut)
                log.debug('supmol_d nbas = %d cGTO = %d', supmol_d.nbas, supmol_d.nao)
        log.timer_debug1('initializing supmol', *cpu0)

        self._cintopt = _vhf.make_cintopt(supmol_sr._atm, supmol_sr._bas,
                                          supmol_sr._env, 'int2e')
        nbas = supmol_sr.nbas
        qindex = np.empty((3,nbas,nbas), dtype=np.int16)
        ao_loc = supmol_sr.ao_loc
        with supmol_sr.with_integral_screen(self.direct_scf_tol**2):
            libpbc.PBCVHFnr_int2e_q_cond(
                libpbc.int2e_sph, self._cintopt,
                qindex.ctypes.data_as(ctypes.c_void_p),
                ao_loc.ctypes.data_as(ctypes.c_void_p),
                supmol_sr._atm.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(supmol_sr.natm),
                supmol_sr._bas.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(supmol_sr.nbas),
                supmol_sr._env.ctypes.data_as(ctypes.c_void_p))
        libpbc.PBCVHFnr_sindex(
            qindex[2:].ctypes.data_as(ctypes.c_void_p),
            supmol_sr._atm.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(supmol_sr.natm),
            supmol_sr._bas.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(supmol_sr.nbas),
            supmol_sr._env.ctypes.data_as(ctypes.c_void_p))

        if self.exclude_dd_block:
            # Remove the smooth-smooth basis block.
            smooth_idx = supmol_sr.bas_type_to_indices(ft_ao.SMOOTH_BASIS)
            qindex[0,smooth_idx[:,None], smooth_idx] = INDEX_MIN

        self.qindex = qindex

        log.timer('initializing qindex', *cpu0)
        return self

    @property
    def qindex(self):
        if self._qindex is None or isinstance(self._qindex, np.ndarray):
            return self._qindex
        return self._qindex['qindex'][:]

    @qindex.setter
    def qindex(self, x):
        if x.size < self.max_memory*.2e6:
            self._qindex = x
        else:
            self._qindex = lib.H5TmpFile()
            self._qindex['qindex'] = x
            self._qindex.flush()

    def _sort_qcond_cell0(self, seg_loc, seg2sh, nbasp, qindex):
        '''Sort qcond for ij-pair in cell0 so that loops in PBCVHF_direct_drv
        can be "break" early
        '''
        qcell0_ijij = _qcond_cell0_abstract(qindex[0], seg_loc, seg2sh, nbasp)
        qcell0_iijj = _qcond_cell0_abstract(qindex[1], seg_loc, seg2sh, nbasp)
        idx = np.asarray(qcell0_ijij.ravel().argsort(kind='stable'), np.int32)[::-1]
        jsh_idx, ish_idx = divmod(idx, nbasp)
        ish_idx = np.asarray(ish_idx, dtype=np.int32, order='C')
        jsh_idx = np.asarray(jsh_idx, dtype=np.int32, order='C')
        return qcell0_ijij, qcell0_iijj, ish_idx, jsh_idx

    def _get_jk_sr(self, dm_kpts, hermi=1, kpts=None, kpts_band=None,
                   with_j=True, with_k=True, omega=None, exxdiv=None):
        if omega is not None:  # J/K for RSH functionals
            raise NotImplementedError

        cpu0 = logger.process_clock(), logger.perf_counter()
        log = logger.new_logger(self)

        comp = 1
        nkpts = kpts.shape[0]
        supmol = self.supmol_sr
        cell = self.cell
        rs_cell = self.rs_cell
        nao = cell.nao
        bvk_ncells, rs_nbas, nimgs = supmol.bas_mask.shape

        if dm_kpts.ndim != 4:
            dm = dm_kpts.reshape(-1, nkpts, nao, nao)
        else:
            dm = dm_kpts
        n_dm = dm.shape[0]

        cutoff = int(np.log(self.direct_scf_tol) * LOG_ADJUST)

        self._sr_without_dddd = (
            self.allow_drv_nodddd and
            not self.exclude_dd_block and
            self.cell_d.nbas > 0 and
            self.has_long_range() and
            (cell.dimension == 3 or cell.low_dim_ft_type != 'inf_vacuum'))

        if self._sr_without_dddd:
            # To exclude (dd|dd) block, diffused shell needs to be independent
            # segments in PBCVHF_direct_drv1. Decontract the segments in rs-shell.
            # map ao indices in cell to ao indices in rs_cell
            log.debug1('get_jk_sr with PBCVHF_direct_drv_nodddd')
            drv = libpbc.PBCVHF_direct_drv_nodddd
            ao_map = lib.locs_to_indices(cell.ao_loc, rs_cell.bas_map)
            rs_nao = ao_map.size
            rs_dm = np.empty((n_dm,nkpts*rs_nao,rs_nao), dtype=dm.dtype)
            idx = np.arange(nkpts)[:,None] * nao + ao_map
            idx = np.asarray(idx, dtype=np.int32).ravel()
            for i in range(n_dm):
                lib.take_2d(dm[i].reshape(nkpts*nao,nao), idx, ao_map, out=rs_dm[i])
            dm = rs_dm.reshape(n_dm,nkpts,rs_nao,rs_nao)
            nbasp = rs_cell.nbas
            cell0_ao_loc = rs_cell.ao_loc

            # uncontracted (local, diffused) basis segments in bvk-cell
            seg_loc = lib.locs_to_indices(
                supmol.seg_loc, np.arange(supmol.seg_loc.size-1))
            seg_loc = np.append(seg_loc, supmol.seg_loc[-1]).astype(np.int32)
        else:
            drv = libpbc.PBCVHF_direct_drv
            nbasp = cell.nbas  # The number of shells in the primitive cell
            cell0_ao_loc = cell.ao_loc
            seg_loc = supmol.seg_loc

        qindex = self.qindex
        qcell0_ijij, qcell0_iijj, ish_idx, jsh_idx = \
                self._sort_qcond_cell0(seg_loc, supmol.seg2sh, nbasp, qindex)

        weight = 1. / nkpts
        expLk = np.exp(1j*np.dot(supmol.bvkmesh_Ls, kpts.T))
        # Utilized symmetry sc_dm[R,S] = sc_dm[S-R] = sc_dm[(S-R)%N]
        #:phase = expLk / nkpts**.5
        #:sc_dm = lib.einsum('Rk,nkuv,Sk->nRuSv', phase, sc_dm, phase.conj())
        sc_dm = lib.einsum('k,Sk,nkuv->nSuv', expLk[0]*weight, expLk.conj(), dm)
        dm_translation = k2gamma.double_translation_indices(self.bvk_kmesh).astype(np.int32)
        dm_imag_max = abs(sc_dm.imag).max()
        is_complex_dm = dm_imag_max > 1e-6
        if is_complex_dm:
            if dm_imag_max < 1e-2:
                log.warn('DM in (BvK) cell has small imaginary part.  '
                         'It may be a signal of symmetry broken in k-point symmetry')
            sc_dm = np.vstack([sc_dm.real, sc_dm.imag])
            self.purify = False
        else:
            sc_dm = sc_dm.real
        nao1 = dm.shape[-1]
        sc_dm = np.asarray(sc_dm.reshape(-1, bvk_ncells, nao1, nao1), order='C')
        n_sc_dm = sc_dm.shape[0]

        # * sparse_ao_loc has dimension (Nk,nbas), corresponding to the
        # bvkcell with all basis
        sparse_ao_loc = nao1 * np.arange(bvk_ncells)[:,None] + cell0_ao_loc[:-1]
        sparse_ao_loc = np.append(sparse_ao_loc.ravel(), nao1 * bvk_ncells)
        dm_cond = [lib.condense('NP_absmax', d, sparse_ao_loc, cell0_ao_loc)
                   for d in sc_dm.reshape(n_sc_dm, bvk_ncells*nao1, nao1)]
        dm_cond = np.max(dm_cond, axis=0)
        dm_cond[dm_cond < 1e-100] = 1e-100
        dmindex = np.log(dm_cond)
        dmindex *= LOG_ADJUST
        dmindex = np.asarray(dmindex, order='C', dtype=np.int16)
        dmindex = dmindex.reshape(bvk_ncells, nbasp, nbasp)
        dm_cond = None

        bvk_nbas = nbasp * bvk_ncells
        shls_slice = (0, nbasp, 0, bvk_nbas, 0, bvk_nbas, 0, bvk_nbas)

        cache_size = _get_cache_size(cell, 'int2e_sph')
        cell0_dims = cell0_ao_loc[1:] - cell0_ao_loc[:-1]
        cache_size += int(cell0_dims.max())**4 * comp * 2

        if hermi:
            fdot_suffix = 's2kl'
        else:
            fdot_suffix = 's1'
        nvs = 1
        if with_j and with_k:
            fdot = 'PBCVHF_contract_jk_' + fdot_suffix
            nvs = 2
        elif with_j:
            fdot = 'PBCVHF_contract_j_' + fdot_suffix
        else:  # with_k
            fdot = 'PBCVHF_contract_k_' + fdot_suffix
        vs = np.zeros((nvs, n_sc_dm, nao1, bvk_ncells, nao1))

        if supmol.cart:
            intor = 'PBCint2e_cart'
        else:
            intor = 'PBCint2e_sph'

        drv(getattr(libpbc, fdot), getattr(libpbc, intor),
            vs.ctypes.data_as(ctypes.c_void_p),
            sc_dm.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(vs.size), ctypes.c_int(n_sc_dm),
            ctypes.c_int(bvk_ncells), ctypes.c_int(nimgs), ctypes.c_int(nkpts),
            ctypes.c_int(nbasp), ctypes.c_int(comp),
            seg_loc.ctypes.data_as(ctypes.c_void_p),
            supmol.seg2sh.ctypes.data_as(ctypes.c_void_p),
            cell0_ao_loc.ctypes.data_as(ctypes.c_void_p),
            rs_cell.bas_type.ctypes.data_as(ctypes.c_void_p),
            (ctypes.c_int*8)(*shls_slice),
            dm_translation.ctypes.data_as(ctypes.c_void_p),
            qindex.ctypes.data_as(ctypes.c_void_p),
            dmindex.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(cutoff),
            qcell0_ijij.ctypes.data_as(ctypes.c_void_p),
            qcell0_iijj.ctypes.data_as(ctypes.c_void_p),
            ish_idx.ctypes.data_as(ctypes.c_void_p),
            jsh_idx.ctypes.data_as(ctypes.c_void_p),
            self._cintopt, ctypes.c_int(cache_size),
            supmol._atm.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(supmol.natm),
            supmol._bas.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(supmol.nbas),
            supmol._env.ctypes.data_as(ctypes.c_void_p))

        if is_complex_dm:
            vs = vs[:,:n_dm] + vs[:,n_dm:] * 1j

        if self._sr_without_dddd:
            jdx = np.arange(bvk_ncells)[:,None] * nao + ao_map
            jdx = np.asarray(jdx, dtype=np.int32).ravel()
            rs_vs = vs.reshape(nvs*n_dm, nao1, bvk_ncells*nao1)
            vs = np.zeros((nvs*n_dm, nao, bvk_ncells*nao), dtype=rs_vs.dtype)
            for i in range(nvs*n_dm):
                lib.takebak_2d(vs[i], rs_vs[i], ao_map, jdx, thread_safe=False)
            vs = vs.reshape(nvs,n_dm,nao,bvk_ncells,nao)

        if kpts_band is not None:
            kpts_band = np.reshape(kpts_band, (-1, 3))
            subset_only = intersection(kpts, kpts_band).size == len(kpts_band)
            if not subset_only:
                log.warn('Approximate J/K matrices at kpts_band '
                         'with the bvk-cell derived from kpts')
                expLk = np.exp(1j*np.dot(supmol.bvkmesh_Ls, kpts_band.T))
        vs = lib.einsum('snpRq,Rk->snkpq', vs, expLk)
        vs = np.asarray(vs, order='C')
        log.timer_debug1('short range part vj and vk', *cpu0)
        return vs

    def get_jk(self, dm_kpts, hermi=1, kpts=None, kpts_band=None,
               with_j=True, with_k=True, omega=None, exxdiv=None):
        cell = self.cell
        if omega is not None:  # J/K for RSH functionals
            if omega > 0:  # Long-range part only, call AFTDF
                dfobj = aft.AFTDF(cell, self.kpts)
                ke_cutoff = estimate_ke_cutoff_for_omega(cell, omega)
                dfobj.mesh = cell.cutoff_to_mesh(ke_cutoff)
                return dfobj.get_jk(dm_kpts, hermi, kpts, kpts_band,
                                    with_j, with_k, omega, exxdiv)
            elif omega < 0:  # Short-range part only
                if self.omega is not None and self.omega != omega:
                    raise RuntimeError(f'omega = {omega}, self.omega = {self.omega}')
                raise NotImplementedError

        if self.supmol_sr is None:
            self.build()

        # Does not support to specify arbitrary kpts
        if kpts is not None and abs(kpts-self.kpts).max() > 1e-7:
            raise RuntimeError('kpts error. kpts cannot be modified in RSJK')
        kpts = np.asarray(self.kpts.reshape(-1, 3), order='C')

        mo_coeff = getattr(dm_kpts, 'mo_coeff', None)
        mo_occ = getattr(dm_kpts, 'mo_occ', None)
        dm_kpts = np.asarray(dm_kpts)
        dms = _format_dms(dm_kpts, kpts)

        # compute delta vs if dm is obtained from SCF make_rdm1
        if mo_coeff is not None:
            last_dm, last_vs = self._last_vs
            vs = self._get_jk_sr(dms-last_dm, hermi, kpts, kpts_band,
                                 with_j, with_k, omega, exxdiv)
            vs += last_vs
            self._last_vs = (dms, vs.copy())
            last_dm = last_vs = None

            # dm ~= dm_factor * dm_factor.T
            n_dm, nkpts, nao = dms.shape[:3]
            # mo_coeff, mo_occ are not a list of aligned array if
            # remove_lin_dep was applied to scf object
            if dm_kpts.ndim == 4:  # KUHF
                nocc = max(max(np.count_nonzero(x > 0) for x in z) for z in mo_occ)
                dm_factor = [[x[:,:nocc] for x in mo] for mo in mo_coeff]
                occs = [[x[:nocc] for x in z] for z in mo_occ]
            else:  # KRHF
                nocc = max(np.count_nonzero(x > 0) for x in mo_occ)
                dm_factor = [[mo[:,:nocc] for mo in mo_coeff]]
                occs = [[x[:nocc] for x in mo_occ]]
            dm_factor = np.array(dm_factor, dtype=np.complex128, order='C')
            dm_factor *= np.sqrt(np.array(occs, dtype=np.double))[:,:,None]
        else:
            vs = self._get_jk_sr(dms, hermi, kpts, kpts_band,
                                 with_j, with_k, omega, exxdiv)
            dm_factor = None
        dms = lib.tag_array(dms, dm_factor=dm_factor)

        if with_j and with_k:
            vj, vk = vs
        elif with_j:
            vj, vk = vs[0], None
        else:
            vj, vk = None, vs[0]

        if self.purify and kpts_band is None:
            phase = np.exp(1j*np.dot(self.supmol_sr.bvkmesh_Ls, kpts.T))
            phase /= np.sqrt(len(kpts))
        else:
            phase = None

        if with_j:
            if self.has_long_range():
                vj += self._get_vj_lr(dms, hermi, kpts, kpts_band)
            if hermi:
                vj = (vj + vj.conj().transpose(0,1,3,2)) * .5
            vj = _purify(vj, phase)
            vj = _format_jks(vj, dm_kpts, kpts_band, kpts)
            if is_zero(kpts) and dm_kpts.dtype == np.double:
                vj = vj.real.copy()

        if with_k:
            if self.has_long_range():
                approx_vk_lr = dm_factor is None and self.approx_vk_lr_missing_mo
                if not approx_vk_lr:
                    vk += self._get_vk_lr(dms, hermi, kpts, kpts_band, exxdiv)
                else:
                    mesh1 = np.array(self.mesh)//3*2 + 1
                    logger.debug(self, 'Approximate lr_k with mesh %s', mesh1)
                    with lib.temporary_env(self, mesh=mesh1):
                        vk += self._get_vk_lr(dms, hermi, kpts, kpts_band, exxdiv)
                    self.approx_vk_lr_missing_mo = False
            if hermi:
                vk = (vk + vk.conj().transpose(0,1,3,2)) * .5
            vk = _purify(vk, phase)
            vk = _format_jks(vk, dm_kpts, kpts_band, kpts)
            if is_zero(kpts) and dm_kpts.dtype == np.double:
                vk = vk.real.copy()

        return vj, vk

    weighted_coulG = aft.weighted_coulG

    def weighted_coulG_LR(self, kpt=np.zeros(3), exx=False, mesh=None):
        # The long range part Coulomb kernel has to be computed as the
        # difference between coulG(cell.omega) - coulG(self.omega). It allows this
        # module to handle the SR- and regular integrals in the same framework
        return (self.weighted_coulG(kpt, exx, mesh) -
                self.weighted_coulG_SR(kpt, exx, mesh))

    def weighted_coulG_SR(self, kpt=np.zeros(3), exx=False, mesh=None):
        return self.weighted_coulG(kpt, False, mesh, -self.omega)

    def _get_vj_lr(self, dm_kpts, hermi=1, kpts=np.zeros((1,3)), kpts_band=None):
        '''
        Long-range part of J matrix
        '''
        if kpts_band is None:
            return self._get_lr_j_kpts(dm_kpts, hermi, kpts)

        logger.warn(self, 'Approximate kpts_band for vj with k-point projection')
        vj = self._get_lr_j_kpts(dm_kpts, hermi, kpts)
        pk2k = addons._k2k_projection(kpts, kpts_band, self.supmol_ft.bvkmesh_Ls)
        return lib.einsum('nkpq,kh->nhpq', vj, pk2k)

    def _get_lr_j_kpts(self, dm_kpts, hermi=1, kpts=np.zeros((1,3))):
        '''
        Long-range part of J matrix
        '''
        if len(kpts) == 1 and not is_zero(kpts):
            raise NotImplementedError('Single k-point get-j')

        cpu0 = logger.process_clock(), logger.perf_counter()
        log = logger.new_logger(self)
        cell = self.cell
        rs_cell = self.rs_cell
        if self.exclude_dd_block or self._sr_without_dddd:
            cell_d = self.cell_d
            naod = cell_d.nao
            ngrids_d = np.prod(cell_d.mesh)
        else:
            cell_d = None
            naod = ngrids_d = 0
        kpts = np.asarray(kpts.reshape(-1, 3), order='C')
        dms = dm_kpts
        n_dm, nkpts, nao = dms.shape[:3]
        mesh = self.mesh
        ngrids = np.prod(mesh)

        is_real = is_zero(kpts) and dms.dtype == np.double
        if is_real:
            vj_kpts = np.zeros((n_dm,nkpts,nao,nao))
        else:
            vj_kpts = np.zeros((n_dm,nkpts,nao,nao), dtype=np.complex128)

        # TODO: aosym == 's2'
        aosym = 's1'
        ft_kern = self.supmol_ft.gen_ft_kernel(
            aosym, return_complex=False, kpts=kpts, verbose=log)
        Gv, Gvbase, kws = cell.get_Gv_weights(mesh)
        gxyz = lib.cartesian_prod([np.arange(len(x)) for x in Gvbase])

        kpt_allow = np.zeros(3)
        coulG = self.weighted_coulG(kpt_allow, False, mesh)

        if (cell.dimension == 3 or
            (cell.dimension == 2 and cell.low_dim_ft_type != 'inf_vacuum')):
            G0_idx = 0  # due to np.fft.fftfreq convention
            # G=0 associated to 2e integrals in real-space
            coulG_SR_at_G0 = np.pi/self.omega**2
            # For cell.dimension = 2, coulG is computed with truncated Coulomb
            # interactions. The 3d coulG_SR below is to remove the analytical
            # SR from get_jk_sr (which was computed with full Coulomb) then to
            # add truncated Coulomb for AFT part.
            with lib.temporary_env(cell, dimension=3):
                coulG_SR = self.weighted_coulG_SR(kpt_allow, False, mesh)
            coulG_SR[G0_idx] += coulG_SR_at_G0 * kws
        else:
            coulG_SR = self.weighted_coulG_SR(kpt_allow, False, mesh)
            coulG_SR_at_G0 = None

        if naod > 0:
            smooth_bas_mask = rs_cell.bas_type == ft_ao.SMOOTH_BASIS
            smooth_bas_idx = rs_cell.bas_map[smooth_bas_mask]
            ao_d_idx = rs_cell.get_ao_indices(smooth_bas_idx, cell.ao_loc)

        mem_now = lib.current_memory()[0]
        max_memory = self.max_memory - mem_now
        log.debug1('max_memory = %d MB (%d in use)', max_memory+mem_now, mem_now)
        Gblksize = max(24, int(max_memory*.8e6/16/(nao**2+naod**2)/(nkpts+1))//8*8)
        Gblksize = min(Gblksize, ngrids)
        log.debug1('Gblksize = %d', Gblksize)

        if not self.exclude_dd_block or naod == 0:
            log.debug1('get_lr_j_kpts with aft_aopair')
            # Long-range part is calculated as the difference
            # coulG(cell.omega) - coulG(self.omega) . It can support both regular
            # integrals and LR integrals.
            coulG_LR = coulG - coulG_SR
            buf = np.empty(nkpts*Gblksize*nao**2*2)
            for p0, p1 in lib.prange(0, ngrids, Gblksize):
                Gpq = ft_kern(Gv[p0:p1], gxyz[p0:p1], Gvbase, kpt_allow, out=buf)
                aft_jk._update_vj_(vj_kpts, Gpq, dms, coulG_LR[p0:p1])
            Gpq = buf = None

            if self._sr_without_dddd and naod > 0:
                log.debug1('get_lr_j_kpts dd block with mesh %s', cell_d.mesh)
                if cell.dimension < 2:
                    raise NotImplementedError
                if cell.dimension == 2 and cell.low_dim_ft_type == 'inf_vacuum':
                    raise NotImplementedError
                aoR_ks, aoI_ks = rsdf_builder._eval_gto(cell_d, cell_d.mesh, kpts)

                # rho = einsum('nkji,kig,kjg->ng', dm, ao.conj(), ao)
                rho = np.zeros((n_dm, ngrids_d))
                tmpR = np.empty((naod, ngrids_d))
                tmpI = np.empty((naod, ngrids_d))
                dm_dd = dms[:,:,ao_d_idx[:,None],ao_d_idx]
                dmR_dd = np.asarray(dm_dd.real, order='C')
                dmI_dd = np.asarray(dm_dd.imag, order='C')
                # vG = einsum('ij,gji->g', dm_dd[k], aoao[k]) * coulG
                for i in range(n_dm):
                    for k in range(nkpts):
                        zdotNN(dmR_dd[i,k].T, dmI_dd[i,k].T,
                               aoR_ks[k], aoI_ks[k], 1, tmpR, tmpI)
                        rho[i] += np.einsum('ig,ig->g', aoR_ks[k], tmpR)
                        rho[i] += np.einsum('ig,ig->g', aoI_ks[k], tmpI)

                if coulG_SR_at_G0 is not None:
                    with lib.temporary_env(cell_d, dimension=3):
                        coulG_SR = pbctools.get_coulG(cell_d, omega=-self.omega)
                    coulG_SR[G0_idx] += coulG_SR_at_G0
                else:
                    coulG_SR = pbctools.get_coulG(cell_d, omega=-self.omega)
                coulG_SR *= cell.vol / ngrids_d
                vG = pbctools.fft(rho, cell_d.mesh) * coulG_SR
                vR = pbctools.ifft(vG, cell_d.mesh).real

                nband = nkpts
                vjR_dd = np.empty((naod,naod))
                vjI_dd = np.empty((naod,naod))
                for i in range(n_dm):
                    for k in range(nband):
                        aowR = np.einsum('xi,x->xi', aoR_ks[k].T, vR[i])
                        aowI = np.einsum('xi,x->xi', aoI_ks[k].T, vR[i])
                        zdotCN(aoR_ks[k], aoI_ks[k], aowR, aowI, 1, vjR_dd, vjI_dd)
                        if is_real:
                            vj = vjR_dd
                        else:
                            vj = vjR_dd + vjI_dd*1j
                        lib.takebak_2d(vj_kpts[i,k], vj, ao_d_idx, ao_d_idx,
                                       thread_safe=False)

        elif naod > 0 and ngrids < ngrids_d:
            # Prefer AFTDF for everything otherwise cell_d.mesh have to be used
            # for AFTDF
            log.debug1('get_lr_j_kpts dd block cached aft_aopair_dd')
            ft_kern_dd = self.supmol_d.gen_ft_kernel(
                aosym, return_complex=False, kpts=kpts, verbose=log)

            def merge_dd(Gpq, p0, p1):
                '''Merge diffused basis block into ao-pair tensor inplace'''
                GpqR, GpqI = Gpq
                pqG_ddR, pqG_ddI = ft_kern_dd(Gv[p0:p1], gxyz[p0:p1], Gvbase,
                                              kpt_allow, kpts)
                # Gpq should be an array of (nkpts,ni,nj,ngrids) in C order
                if not GpqR[0].flags.c_contiguous:
                    assert GpqR[0].strides[0] == 8  # stride for grids
                for k in range(nkpts):
                    libpbc.PBC_ft_fuse_dd_s1(
                        GpqR[k].ctypes.data_as(ctypes.c_void_p),
                        GpqI[k].ctypes.data_as(ctypes.c_void_p),
                        pqG_ddR[k].ctypes.data_as(ctypes.c_void_p),
                        pqG_ddI[k].ctypes.data_as(ctypes.c_void_p),
                        ao_d_idx.ctypes.data_as(ctypes.c_void_p),
                        ctypes.c_int(nao), ctypes.c_int(naod), ctypes.c_int(p1-p0))
                return (GpqR, GpqI)

            buf = np.empty(nkpts*Gblksize*nao**2*2)
            for p0, p1 in lib.prange(0, ngrids, Gblksize):
                Gpq = ft_kern(Gv[p0:p1], gxyz[p0:p1], Gvbase, kpt_allow, out=buf)
                aft_jk._update_vj_(vj_kpts, Gpq, dms, -coulG_SR[p0:p1])
                Gpq = merge_dd(Gpq, p0, p1)
                aft_jk._update_vj_(vj_kpts, Gpq, dms, coulG[p0:p1])
            Gpq = buf = None

        elif naod > 0:
            log.debug1('get_lr_j_kpts dd block cached fft_aopair_dd')
            aoR_ks, aoI_ks = rsdf_builder._eval_gto(cell_d, mesh, kpts)

            # rho = einsum('nkji,kig,kjg->ng', dm, ao.conj(), ao)
            rho = np.zeros((n_dm, ngrids))
            tmpR = np.empty((naod, ngrids))
            tmpI = np.empty((naod, ngrids))
            dm_dd = dms[:,:,ao_d_idx[:,None],ao_d_idx]
            dmR_dd = np.asarray(dm_dd.real, order='C')
            dmI_dd = np.asarray(dm_dd.imag, order='C')
            # vG = einsum('ij,gji->g', dm_dd[k], aoao[k]) * coulG
            for i in range(n_dm):
                for k in range(nkpts):
                    zdotNN(dmR_dd[i,k].T, dmI_dd[i,k].T, aoR_ks[k], aoI_ks[k], 1, tmpR, tmpI)
                    rho[i] += np.einsum('ig,ig->g', aoR_ks[k], tmpR)
                    rho[i] += np.einsum('ig,ig->g', aoI_ks[k], tmpI)
            vG_dd = pbctools.ifft(rho, mesh) * cell.vol * coulG
            tmpR = tmpI = dmR_dd = dmI_dd = None
            cpu1 = log.timer_debug1('get_lr_j_kpts dd block', *cpu0)

            mem_now = lib.current_memory()[0]
            max_memory = self.max_memory - mem_now
            log.debug1('max_memory = %d MB (%d in use)', max_memory+mem_now, mem_now)
            Gblksize = min(Gblksize, int(max_memory*.8e6/16/(nao**2+naod**2)/(nkpts+1))//8*8)
            log.debug1('Gblksize = %d', Gblksize)
            buf = np.empty(nkpts*Gblksize*nao**2*2)
            for p0, p1 in lib.prange(0, ngrids, Gblksize):
                Gpq = ft_kern(Gv[p0:p1], gxyz[p0:p1], Gvbase, kpt_allow, out=buf)
                #: aft_jk._update_vj_(vj_kpts, aoaoks, dms, coulG[p0:p1], 1)
                #: aft_jk._update_vj_(vj_kpts, aoaoks, dms, coulG_SR[p0:p1], -1)
                GpqR, GpqI = Gpq
                for i in range(n_dm):
                    rhoR = np.einsum('kij,kgij->g', dms[i].real, GpqR)
                    rhoI = -np.einsum('kij,kgij->g', dms[i].real, GpqI)
                    if not is_real:
                        rhoR += np.einsum('kij,kgij->g', dms[i].imag, GpqI)
                        rhoI += np.einsum('kij,kgij->g', dms[i].imag, GpqR)
                    rho = rhoR + rhoI * 1j
                    vG = vG_dd[i,p0:p1]
                    # Update vG_dd inplace to include rho contributions
                    vG += coulG[p0:p1] * rho
                    vG_SR = coulG_SR[p0:p1] * rho
                    # vG_LR contains full vG of dd-block and vG_LR of rest blocks
                    vG_LR = vG - vG_SR
                    vj_kpts[i].real += np.einsum('g,kgij->kij', vG_LR.real, GpqR)
                    vj_kpts[i].real -= np.einsum('g,kgij->kij', vG_LR.imag, GpqI)
                    if not is_real:
                        vj_kpts[i].imag += np.einsum('g,kgij->kij', vG_LR.real, GpqI)
                        vj_kpts[i].imag += np.einsum('g,kgij->kij', vG_LR.imag, GpqR)
                Gpq = None
            log.timer_debug1('get_lr_j_kpts ft_aopair', *cpu1)

            vR = pbctools.fft(vG_dd, mesh).real * (cell.vol/ngrids)
            vjR_dd = np.empty((naod, naod))
            vjI_dd = np.empty((naod, naod))
            for i in range(n_dm):
                for k in range(nkpts):
                    tmpR = aoR_ks[k] * vR[i]
                    tmpI = aoI_ks[k] * vR[i]
                    zdotCN(aoR_ks[k], aoI_ks[k], tmpR.T, tmpI.T, 1, vjR_dd, vjI_dd)
                    if is_real:
                        vj_dd = vjR_dd
                    else:
                        vj_dd = vjR_dd + vjI_dd * 1j
                    lib.takebak_2d(vj_kpts[i,k], vj_dd, ao_d_idx, ao_d_idx,
                                   thread_safe=False)

        else:
            raise RuntimeError(f'exclude_dd_block={self.exclude_dd_block} '
                               f'sr_without_dddd={self._sr_without_dddd} naod={naod}')

        if nkpts > 1:
            vj_kpts *= 1./nkpts
        log.timer_debug1('get_lr_j_kpts', *cpu0)
        return vj_kpts

    def _get_vk_lr(self, dm_kpts, hermi=1, kpts=np.zeros((1,3)), kpts_band=None,
                   exxdiv=None):
        '''
        Long-range part of K matrix
        '''
        if kpts_band is None:
            return self._get_lr_k_kpts(dm_kpts, hermi, kpts, exxdiv)

        # Note: Errors in k2k-projection for vj is relatively small.
        # k2k-projection for vk has significant finite-size errors.
        logger.warn(self, 'Approximate kpts_band for vk with k-point projection')
        vk = self._get_lr_k_kpts(dm_kpts, hermi, kpts, exxdiv=None)
        pk2k = addons._k2k_projection(kpts, kpts_band, self.supmol_ft.bvkmesh_Ls)
        vk = lib.einsum('nkpq,kh->nhpq', vk, pk2k)
        if exxdiv == 'ewald':
            _ewald_exxdiv_for_G0(self.cell, kpts, dm_kpts, vk, kpts_band)
        return vk

    def _get_lr_k_kpts(self, dm_kpts, hermi=1, kpts=np.zeros((1,3)), exxdiv=None):
        '''
        Long-range part of K matrix
        '''
        cpu0 = cpu1 = logger.process_clock(), logger.perf_counter()
        log = logger.new_logger(self)
        cell = self.cell
        rs_cell = self.rs_cell
        if self.exclude_dd_block or self._sr_without_dddd:
            cell_d = self.cell_d
            naod = cell_d.nao
            ngrids_d = np.prod(cell_d.mesh)
        else:
            cell_d = None
            naod = ngrids_d = 0

        mesh = self.mesh
        ngrids = np.prod(mesh)
        dms = dm_kpts
        n_dm, nkpts, nao = dms.shape[:3]
        vkR = np.zeros((n_dm,nkpts,nao,nao))
        vkI = np.zeros((n_dm,nkpts,nao,nao))
        vk = [vkR, vkI]
        weight = 1. / nkpts

        # Test if vk[k] == vk[k_conj].conj()
        t_rev_pairs = group_by_conj_pairs(cell, kpts, return_kpts_pairs=False)
        try:
            t_rev_pairs = np.asarray(t_rev_pairs, dtype=np.int32, order='F')
        except TypeError:
            t_rev_pairs = [[k, k] if k_conj is None else [k, k_conj]
                           for k, k_conj in t_rev_pairs]
            t_rev_pairs = np.asarray(t_rev_pairs, dtype=np.int32, order='F')
        log.debug1('Num kpts conj_pairs %d', len(t_rev_pairs))
        time_reversal_symmetry = self.time_reversal_symmetry
        if time_reversal_symmetry:
            for k, k_conj in t_rev_pairs:
                if k != k_conj and abs(dms[:,k_conj] - dms[:,k].conj()).max() > 1e-6:
                    time_reversal_symmetry = False
                    log.debug2('Disable time_reversal_symmetry')
                    break

        # TODO: aosym == 's2'
        aosym = 's1'
        if time_reversal_symmetry:
            k_to_compute = np.zeros(nkpts, dtype=np.int8)
            k_to_compute[t_rev_pairs[:,0]] = 1
        else:
            k_to_compute = np.ones(nkpts, dtype=np.int8)
            t_rev_pairs = None

        dm_factor = getattr(dm_kpts, 'dm_factor', None)
        contract_mo_early = False
        if dm_factor is None:
            dmsR = np.asarray(dms.real, order='C')
            dmsI = np.asarray(dms.imag, order='C')
            dm = [dmsR, dmsI]
            dm_factor = None
            if np.count_nonzero(k_to_compute) >= 2 * lib.num_threads():
                update_vk = aft_jk._update_vk1_
            else:
                update_vk = aft_jk._update_vk_
        else:
            # dm ~= dm_factor * dm_factor.T
            nocc = dm_factor.shape[-1]
            if nocc == 0:
                return vkR

            bvk_ncells, rs_nbas, nimgs = self.supmol_ft.bas_mask.shape
            s_nao = self.supmol_ft.nao
            contract_mo_early = (time_reversal_symmetry and naod == 0 and
                                 bvk_ncells*nao*6 > s_nao*nocc*n_dm)
            log.debug2('time_reversal_symmetry = %s bvk_ncells = %d '
                       's_nao = %d nocc = %d n_dm = %d',
                       time_reversal_symmetry, bvk_ncells, s_nao, nocc, n_dm)
            log.debug2('Use algorithm contract_mo_early = %s', contract_mo_early)
            if contract_mo_early:
                bvk_kmesh = self.bvk_kmesh
                rcut = ft_ao.estimate_rcut(cell)
                supmol = ft_ao.ExtendedMole.from_cell(cell, bvk_kmesh, rcut.max())
                supmol = supmol.strip_basis(rcut)
                s_nao = supmol.nao
                moR, moI = aft_jk._mo_k2gamma(supmol, dm_factor, kpts, t_rev_pairs)
                if abs(moI).max() < 1e-5:
                    dm = [moR, None]
                    ft_kern = aft_jk._gen_ft_kernel_fake_gamma(cell, supmol, aosym)
                    update_vk = aft_jk._update_vk_fake_gamma
                else:
                    contract_mo_early = False
                moR = moI = None

            if not contract_mo_early:
                dm = [np.asarray(dm_factor.real, order='C'),
                      np.asarray(dm_factor.imag, order='C')]
                if np.count_nonzero(k_to_compute) >= 2 * lib.num_threads():
                    update_vk = aft_jk._update_vk1_dmf
                else:
                    update_vk = aft_jk._update_vk_dmf
        log.debug2('set update_vk to %s', update_vk)

        if not contract_mo_early:
            ft_kern = self.supmol_ft.gen_ft_kernel(
                aosym, return_complex=False, kpts=kpts, verbose=log)

        Gv, Gvbase, kws = cell.get_Gv_weights(mesh)
        Gv = np.asarray(Gv, order='F')
        gxyz = lib.cartesian_prod([np.arange(len(x)) for x in Gvbase])
        G0_idx = 0
        if (cell.dimension == 3 or
            (cell.dimension == 2 and cell.low_dim_ft_type != 'inf_vacuum')):
            coulG_SR_at_G0 = np.pi/self.omega**2 * kws
        else:
            coulG_SR_at_G0 = None

        if naod > 0:
            smooth_bas_mask = rs_cell.bas_type == ft_ao.SMOOTH_BASIS
            smooth_bas_idx = rs_cell.bas_map[smooth_bas_mask]
            ao_d_idx = rs_cell.get_ao_indices(smooth_bas_idx, cell.ao_loc)

        mem_now = lib.current_memory()[0]
        max_memory = max(2000, (self.max_memory - mem_now))
        log.debug1('max_memory = %d MB (%d in use)', max_memory+mem_now, mem_now)

        if self.exclude_dd_block and not self._sr_without_dddd and naod > 0:
            cache_size = (naod*(naod+1)*ngrids*(nkpts+1))*16e-6
            log.debug1('naod = %d cache_size = %d', naod, cache_size)

            # fft_aopair_dd seems less efficient than aft_aopair_dd
            if 0 and max_memory * .5 > cache_size and cell.dimension >= 2:
                from pyscf.pbc.dft.multigrid import _take_5d
                mesh_d = cell_d.mesh
                log.debug1('merge_dd with cached fft_aopair_dd')
                aoR_ks, aoI_ks = rsdf_builder._eval_gto(cell_d, mesh_d, kpts)
                coords = cell_d.get_uniform_grids(mesh_d)
                max_memory -= cache_size
                gx = np.fft.fftfreq(mesh[0], 1./mesh[0]).astype(np.int32)
                gy = np.fft.fftfreq(mesh[1], 1./mesh[1]).astype(np.int32)
                gz = np.fft.fftfreq(mesh[2], 1./mesh[2]).astype(np.int32)

                def fft_aopair_dd(ki, kj, expmikr):
                    # einsum('g,ig,jg->ijg', expmikr, ao_ki.conj(), ao_kj)
                    pqG_ddR = np.empty((naod**2, ngrids_d))
                    pqG_ddI = np.empty((naod**2, ngrids_d))
                    expmikrR, expmikrI = expmikr
                    libpbc.PBC_zjoin_fCN_s1(
                        pqG_ddR.ctypes.data_as(ctypes.c_void_p),
                        pqG_ddI.ctypes.data_as(ctypes.c_void_p),
                        expmikrR.ctypes.data_as(ctypes.c_void_p),
                        expmikrI.ctypes.data_as(ctypes.c_void_p),
                        aoR_ks[ki].ctypes.data_as(ctypes.c_void_p),
                        aoI_ks[ki].ctypes.data_as(ctypes.c_void_p),
                        aoR_ks[kj].ctypes.data_as(ctypes.c_void_p),
                        aoI_ks[kj].ctypes.data_as(ctypes.c_void_p),
                        ctypes.c_int(naod), ctypes.c_int(naod), ctypes.c_int(ngrids_d))
                    pqG_dd = np.empty((naod**2, ngrids_d), dtype=np.complex128)
                    pqG_dd.real = pqG_ddR
                    pqG_dd.imag = pqG_ddI
                    pqG_ddR = pqG_ddI = None
                    pqG_dd *= cell.vol / ngrids_d
                    pqG_dd = pbctools.fft(pqG_dd, mesh_d).reshape(naod, naod, *mesh_d)
                    pqG_dd = _take_5d(pqG_dd, (None, None, gx, gy, gz))
                    return np.asarray(pqG_dd.reshape(naod, naod, ngrids), order='C')

                cache = {}
                def merge_dd(Gpq, p0, p1, ki_lst, kj_lst):
                    '''Merge diffused basis block into ao-pair tensor inplace'''
                    expmikr = np.exp(-1j * np.dot(coords, kpts[kj_lst[0]]-kpts[ki_lst[0]]))
                    expmikrR = np.asarray(expmikr.real, order='C')
                    expmikrI = np.asarray(expmikr.imag, order='C')
                    GpqR, GpqI = Gpq
                    # Gpq should be an array of (nkpts,ni,nj,ngrids) in C order
                    if not GpqR[0].flags.c_contiguous:
                        assert GpqR[0].strides[0] == 8  # stride for grids
                    cpu0 = logger.process_clock(), logger.perf_counter()
                    for k, (ki, kj) in enumerate(zip(ki_lst, kj_lst)):
                        if (ki, kj) not in cache:
                            log.debug3('cache dd block (%d, %d)', ki, kj)
                            cache[ki, kj] = fft_aopair_dd(ki, kj, (expmikrR, expmikrI))

                        pqG_dd = cache[ki, kj]
                        libpbc.PBC_ft_zfuse_dd_s1(
                            GpqR[kj].ctypes.data_as(ctypes.c_void_p),
                            GpqI[kj].ctypes.data_as(ctypes.c_void_p),
                            pqG_dd.ctypes.data_as(ctypes.c_void_p),
                            ao_d_idx.ctypes.data_as(ctypes.c_void_p),
                            (ctypes.c_int*2)(p0, p1), ctypes.c_int(nao),
                            ctypes.c_int(naod), ctypes.c_int(ngrids))
                    log.timer_debug1('merge_dd', *cpu0)
                    return (GpqR, GpqI)

                cpu1 = log.timer_debug1('get_lr_k_kpts initializing dd block', *cpu1)

            else:
                log.debug1('merge_dd with aft_aopair_dd')
                ft_kern_dd = self.supmol_d.gen_ft_kernel(
                    aosym, return_complex=False, kpts=kpts, verbose=log)

                buf1 = None
                def merge_dd(Gpq, p0, p1, ki_lst, kj_lst):
                    '''Merge diffused basis block into ao-pair tensor inplace'''
                    kpt = kpts[kj_lst[0]] - kpts[ki_lst[0]]
                    GpqR, GpqI = Gpq
                    pqG_ddR, pqG_ddI = ft_kern_dd(Gv[p0:p1], gxyz[p0:p1], Gvbase,
                                                  kpt, out=buf1)
                    # Gpq should be an array of (nkpts,ni,nj,ngrids) in C order
                    if not GpqR[0].flags.c_contiguous:
                        assert GpqR[0].strides[0] == 8  # stride for grids
                    for k in range(nkpts):
                        libpbc.PBC_ft_fuse_dd_s1(
                            GpqR[k].ctypes.data_as(ctypes.c_void_p),
                            GpqI[k].ctypes.data_as(ctypes.c_void_p),
                            pqG_ddR[k].ctypes.data_as(ctypes.c_void_p),
                            pqG_ddI[k].ctypes.data_as(ctypes.c_void_p),
                            ao_d_idx.ctypes.data_as(ctypes.c_void_p),
                            ctypes.c_int(nao), ctypes.c_int(naod), ctypes.c_int(p1-p0))
                    return (GpqR, GpqI)

        if contract_mo_early:
            Gblksize = max(24, int((max_memory*1e6/16-nkpts*nao**2*3)/
                                   (nao*s_nao+nao*nkpts*nocc))//8*8)
            Gblksize = min(Gblksize, ngrids, 200000)
            log.debug1('Gblksize = %d', Gblksize)
            buf = np.empty(Gblksize*s_nao*nao*2)
        else:
            Gblksize = max(24, int(max_memory*.8e6/16/(nao**2+naod**2)/(nkpts+3))//8*8)
            Gblksize = min(Gblksize, ngrids, 200000)
            log.debug1('Gblksize = %d', Gblksize)
            buf = np.empty(nkpts*Gblksize*nao**2*2)
            if naod > 0:
                buf1 = np.empty(nkpts*Gblksize*naod**2*2)

        for group_id, (kpt, ki_idx, kj_idx, self_conj) \
                in enumerate(kk_adapted_iter(cell, kpts)):
            coulG = self.weighted_coulG(kpt, exxdiv, mesh)
            if coulG_SR_at_G0 is not None:
                # For cell.dimension = 2, coulG is computed with truncated coulomb
                # interactions. The 3d coulG_SR below is to remove the analytical
                # SR from get_jk_sr (which was computed with full Coulomb) then
                # add the truncated Coulomb for AFT part.
                with lib.temporary_env(cell, dimension=3):
                    coulG_SR = self.weighted_coulG_SR(kpt, False, mesh)
                if is_zero(kpt):
                    coulG_SR[G0_idx] += coulG_SR_at_G0
            else:
                coulG_SR = self.weighted_coulG_SR(kpt, False, mesh)

            if self.exclude_dd_block and not self._sr_without_dddd and naod > 0:
                for p0, p1 in lib.prange(0, ngrids, Gblksize):
                    # C ~ compact basis, D ~ diffused basis
                    # K matrix with coulG_LR:
                    # (CC|CC) (CC|CD) (CC|DC) (CD|CC) (CD|CD) (CD|DC) (DC|CC) (DC|CD) (DC|DC)
                    # K matrix with full coulG:
                    # (CC|DD) (CD|DD) (DC|DD) (DD|CC) (DD|CD) (DD|DC) (DD|DD)
                    log.debug3('update_vk [%s:%s]', p0, p1)
                    Gpq = ft_kern(Gv[p0:p1], gxyz[p0:p1], Gvbase, kpt, out=buf)
                    update_vk(vk, Gpq, dm, coulG_SR[p0:p1] * -weight, ki_idx,
                              kj_idx, not self_conj, k_to_compute, t_rev_pairs)
                    Gpq = merge_dd(Gpq, p0, p1, ki_idx, kj_idx)
                    update_vk(vk, Gpq, dm, coulG[p0:p1] * weight, ki_idx,
                              kj_idx, not self_conj, k_to_compute, t_rev_pairs)
                    Gpq = None
                # clear cache to release memory for merge_dd function
                cache = {}
            else:
                coulG_LR = coulG - coulG_SR
                for p0, p1 in lib.prange(0, ngrids, Gblksize):
                    log.debug3('update_vk [%s:%s]', p0, p1)
                    Gpq = ft_kern(Gv[p0:p1], gxyz[p0:p1], Gvbase, kpt, out=buf)
                    update_vk(vk, Gpq, dm, coulG_LR[p0:p1] * weight, ki_idx,
                              kj_idx, not self_conj, k_to_compute, t_rev_pairs)
                    Gpq = None
            cpu1 = log.timer_debug1(f'ft_aopair group {group_id}', *cpu1)

        if self._sr_without_dddd and naod > 0:
            # (DD|DD) with full coulG, rest terms with coulG_LR
            log.debug1('ft_aopair dd-block for dddd-block ERI')
            if cell.dimension < 2:
                raise NotImplementedError
            if cell.dimension == 2 and cell.low_dim_ft_type == 'inf_vacuum':
                raise NotImplementedError
            vkR_dd = np.zeros((n_dm,nkpts,naod,naod))
            vkI_dd = np.zeros((n_dm,nkpts,naod,naod))
            vk_dd = [vkR_dd, vkI_dd]
            if dm_factor is None:
                dmsR, dmsI = dm
                dmR_dd = np.asarray(dmsR[:,:,ao_d_idx[:,None],ao_d_idx], order='C')
                dmI_dd = np.asarray(dmsI[:,:,ao_d_idx[:,None],ao_d_idx], order='C')
                dm_dd = [dmR_dd, dmI_dd]
            else:
                assert update_vk is not aft_jk._update_vk_fake_gamma
                dmfR, dmfI = dm
                dmfR_dd = np.asarray(dmfR[:,:,ao_d_idx], order='C')
                dmfI_dd = np.asarray(dmfI[:,:,ao_d_idx], order='C')
                dm_dd = [dmfR_dd, dmfI_dd]

            # AFT mesh for ERI is usually smaller than cell_d.mesh
            ke_cutoff = aft.estimate_ke_cutoff(cell_d)
            mesh_d = cell_d.cutoff_to_mesh(ke_cutoff)
            log.debug1('lr_k dd-block ke_cutoff = %s mesh = %s', ke_cutoff, mesh_d)

            ft_kern_dd = self.supmol_d.gen_ft_kernel(
                aosym, return_complex=False, kpts=kpts, verbose=log)
            Gv, Gvbase, kws = cell_d.get_Gv_weights(mesh_d)
            gxyz = lib.cartesian_prod([np.arange(len(x)) for x in Gvbase])
            ngrids_d = len(Gv)

            Gblksize = max(24, int(max_memory*.8e6/16/naod**2/(nkpts+max(nkpts,3)))//8*8)
            Gblksize = min(Gblksize, ngrids_d)
            log.debug1('Gblksize = %d', Gblksize)
            buf = np.empty(nkpts*Gblksize*naod**2*2)
            for group_id, (kpt, ki_idx, kj_idx, self_conj) \
                    in enumerate(kk_adapted_iter(cell, kpts)):
                with lib.temporary_env(cell, dimension=3):
                    coulG_SR = self.weighted_coulG_SR(kpt, False, mesh_d)
                if is_zero(kpt) and coulG_SR_at_G0 is not None:
                    coulG_SR[G0_idx] += coulG_SR_at_G0

                for p0, p1 in lib.prange(0, ngrids_d, Gblksize):
                    Gpq = ft_kern_dd(Gv[p0:p1], gxyz[p0:p1], Gvbase, kpt, out=buf)
                    update_vk(vk_dd, Gpq, dm_dd, coulG_SR[p0:p1] * weight,
                              ki_idx, kj_idx, not self_conj, k_to_compute, t_rev_pairs)
                    Gpq = None
                cpu1 = log.timer_debug1(f'ft_aopair dd-block group {group_id}', *cpu1)

            for i in range(n_dm):
                for k in range(nkpts):
                    lib.takebak_2d(vkR[i,k], vkR_dd[i,k], ao_d_idx, ao_d_idx,
                                   thread_safe=False)
                    lib.takebak_2d(vkI[i,k], vkI_dd[i,k], ao_d_idx, ao_d_idx,
                                   thread_safe=False)

        buf = buf1 = None
        if is_zero(kpts) and not np.iscomplexobj(dm_kpts):
            vk_kpts = vkR
        else:
            vk_kpts = vkR + vkI * 1j

        # Add ewald_exxdiv contribution because G=0 was not included in the
        # non-uniform grids
        if (exxdiv == 'ewald' and
            (cell.dimension < 2 or  # 0D and 1D are computed with inf_vacuum
             (cell.dimension == 2 and cell.low_dim_ft_type == 'inf_vacuum'))):
            _ewald_exxdiv_for_G0(cell, kpts, dms, vk_kpts, kpts)

        if time_reversal_symmetry:
            for k, k_conj in t_rev_pairs:
                if k != k_conj:
                    vk_kpts[:,k_conj] = vk_kpts[:,k].conj()
        log.timer_debug1('get_lr_k_kpts', *cpu0)
        return vk_kpts

    to_gpu = lib.to_gpu

RangeSeparationJKBuilder = RangeSeparatedJKBuilder

def _purify(mat_kpts, phase):
    if phase is None:
        return mat_kpts
    #:mat_bvk = np.einsum('Rk,nkij,Sk->nRSij', phase, mat_kpts, phase.conj())
    #:return np.einsum('Rk,nRSij,Sk->nkij', phase.conj(), mat_bvk.real, phase)
    nkpts = phase.shape[1]
    mat_bvk = lib.einsum('k,Sk,nkuv->nSuv', phase[0], phase.conj(), mat_kpts)
    return lib.einsum('S,Sk,nSuv->nkuv', nkpts*phase[:,0].conj(), phase, mat_bvk.real)


def estimate_rcut(rs_cell, omega, precision=None,
                  exclude_dd_block=True):
    '''Estimate rcut for 2e SR-integrals'''
    if precision is None:
        # Adjust precision a little bit as errors are found slightly larger than cell.precision.
        precision = rs_cell.precision * 1e-1

    rs_cell = rs_cell
    exps, cs = pbcgto.cell._extract_pgto_params(rs_cell, 'min')
    ls = rs_cell._bas[:,gto.ANG_OF]

    exp_min_idx = exps.argmin()
    cost = cs * (.5*abs(omega)*rs_cell.rcut)**ls / (2*exps)**(ls/2+.75)
    ai_idx = ak_idx = cost.argmax()
    compact_mask = rs_cell.bas_type != ft_ao.SMOOTH_BASIS
    compact_idx = np.where(compact_mask)[0]
    if exclude_dd_block and compact_idx.size > 0:
        ak_idx = compact_idx[cost[compact_idx].argmax()]
    logger.debug2(rs_cell, 'ai_idx=%d ak_idx=%d', ai_idx, ak_idx)
    # Case 1: l in cell0, product kl ~ dc, product ij ~ dd and dc
    # This includes interactions (dc|dd)
    ak = exps[ak_idx]
    lk = rs_cell._bas[ak_idx,gto.ANG_OF]
    ck = cs[ak_idx]
    aj = exps
    lj = ls
    cj = cs
    ai = exps[ai_idx]
    li = rs_cell._bas[ai_idx,gto.ANG_OF]
    ci = cs[ai_idx]
    al = exps[exp_min_idx]
    ll = rs_cell._bas[exp_min_idx,gto.ANG_OF]
    cl = cs[exp_min_idx]

    aij = ai + aj
    akl = ak + al
    lij = li + lj
    lkl = lk + ll
    l4 = lij + lkl
    norm_ang = ((2*li+1)*(2*lj+1)*(2*lk+1)*(2*ll+1)/(4*np.pi)**4)**.5
    c1 = ci * cj * ck * cl * norm_ang
    theta = omega**2*aij*akl/(aij*akl + (aij+akl)*omega**2)
    sfac = omega**2*aj*al/(aj*al + (aj+al)*omega**2) / theta
    fl = 2
    fac = 2**(li+lk)*np.pi**2.5*c1 * theta**(l4-.5)
    fac *= 2*np.pi/rs_cell.vol/theta
    fac /= aij**(li+1.5) * akl**(lk+1.5) * aj**lj * al**ll
    fac *= fl / precision

    r0 = rs_cell.rcut
    r0 = (np.log(fac * r0 * (sfac*r0)**(l4-1) + 1.) / (sfac*theta))**.5
    r0 = (np.log(fac * r0 * (sfac*r0)**(l4-1) + 1.) / (sfac*theta))**.5
    rcut = r0

    if exclude_dd_block and 0 < compact_idx.size < rs_cell.nbas:
        # Case 2: l in cell0, product kl ~ dc, product ij ~ cd
        # so as to exclude interaction (dc|dd)
        smooth_mask = ~compact_mask
        ai, li, ci = ak, lk, ck
        aj = exps[smooth_mask]
        lj = ls[smooth_mask]
        cj = cs[smooth_mask]
        aij = ai + aj
        lij = li + lj
        l4 = lij + lkl
        norm_ang = ((2*li+1)*(2*lj+1)*(2*lk+1)*(2*ll+1)/(4*np.pi)**4)**.5
        c1 = ci * cj * ck * cl * norm_ang
        theta = omega**2*aij*akl/(aij*akl + (aij+akl)*omega**2)
        sfac = omega**2*aj*al/(aj*al + (aj+al)*omega**2) / theta
        fl = 2
        fac = 2**(li+lk)*np.pi**2.5*c1 * theta**(l4-.5)
        fac *= 2*np.pi/rs_cell.vol/theta
        fac /= aij**(li+1.5) * akl**(lk+1.5) * aj**lj * al**ll
        fac *= fl / precision

        r0 = rcut[smooth_mask]
        r0 = (np.log(fac * r0 * (sfac*r0)**(l4-1) + 1.) / (sfac*theta))**.5
        r0 = (np.log(fac * r0 * (sfac*r0)**(l4-1) + 1.) / (sfac*theta))**.5
        rcut[smooth_mask] = r0
    return rcut

def _guess_omega(cell, kpts, mesh=None):
    a = cell.lattice_vectors()
    if cell.dimension == 0:
        if mesh is None:
            mesh = cell.mesh
        ke_cutoff = pbctools.mesh_to_cutoff(a, mesh).min()
        return 0, mesh, ke_cutoff

    precision = cell.precision
    nkpts = len(kpts)
    if mesh is None:
        omega_min = OMEGA_MIN
        ke_min = estimate_ke_cutoff_for_omega(cell, omega_min)
        nk = (cell.nao/25 * nkpts)**(1./3)
        ke_cutoff = 50 / (.7+.25*nk+.05*nk**3)
        ke_cutoff = max(ke_cutoff, ke_min)
        # avoid large omega since numerical issues were found in Rys
        # polynomials when computing SR integrals with nroots > 3
        exps = [e for l, e in zip(cell._bas[:,gto.ANG_OF], cell.bas_exps()) if l != 0]
        if exps:
            omega_max = np.hstack(exps).min()**.5 * 2
            ke_max = estimate_ke_cutoff_for_omega(cell, omega_max)
            ke_cutoff = min(ke_cutoff, ke_max)
        mesh = cell.cutoff_to_mesh(ke_cutoff)
    else:
        mesh = np.asarray(mesh)
    ke_cutoff = min(pbctools.mesh_to_cutoff(a, mesh)[:cell.dimension])
    omega = estimate_omega_for_ke_cutoff(cell, ke_cutoff, precision)
    return omega, mesh, ke_cutoff

def estimate_ke_cutoff_for_omega(cell, omega, precision=None):
    '''Energy cutoff for FFTDF to converge attenuated Coulomb in moment space
    '''
    if precision is None:
        precision = cell.precision
    ai = np.hstack(cell.bas_exps()).max()
    theta = 1./(1./ai + omega**-2)
    fac = 32*np.pi**2 * theta / precision
    Ecut = 20.
    Ecut = np.log(fac / (2*Ecut) + 1.) * 2*theta
    Ecut = np.log(fac / (2*Ecut) + 1.) * 2*theta
    return Ecut

def estimate_omega_for_ke_cutoff(cell, ke_cutoff, precision=None):
    '''The minimal omega in attenuated Coulomb given energy cutoff
    '''
    if precision is None:
        precision = cell.precision
#    # estimation based on \int dk 4pi/k^2 exp(-k^2/4omega) sometimes is not
#    # enough to converge the 2-electron integrals. A penalty term here is to
#    # reduce the error in integrals
#    precision *= 1e-2
#    kmax = (ke_cutoff*2)**.5
#    log_rest = np.log(precision / (16*np.pi**2 * kmax**lmax))
#    omega = (-.5 * ke_cutoff / log_rest)**.5
#    return omega

    ai = np.hstack(cell.bas_exps()).max()
    aij = ai * 2
    fac = 32*np.pi**2 / precision
    omega = .3
    theta = 1./(1./ai + omega**-2)
    omega2 = 1./(np.log(fac * theta/ (2*ke_cutoff) + 1.)*2/ke_cutoff - 1./aij)
    if omega2 > 0:
        theta = 1./(1./ai + 1./omega2)
        omega2 = 1./(np.log(fac * theta/ (2*ke_cutoff) + 1.)*2/ke_cutoff - 1./aij)
    omega = max(omega2, 0)**.5
    if omega < OMEGA_MIN:
        logger.warn(cell, 'omega=%g smaller than the required minimal value %g. '
                    'Set omega to %g', omega2, OMEGA_MIN, OMEGA_MIN)
        omega = OMEGA_MIN
    return omega

def _qcond_cell0_abstract(qcond, seg_loc, seg2sh, nbasp):
    '''Find the max qcond for ij pair'''
    # The first shell in each seg_loc[i]:seg_loc[i+1] is inside cell0
    cell0_prim_idx = seg2sh[np.arange(seg_loc[nbasp])]
    qcond_sub = qcond[:,cell0_prim_idx]
    sh_loc = seg2sh[seg_loc]
    nbas_bvk = sh_loc.size - 1
    qtmp = np.empty((nbas_bvk, cell0_prim_idx.size), dtype=qcond.dtype)
    for i, (i0, i1) in enumerate(zip(sh_loc[:-1], sh_loc[1:])):
        if i0 != i1:
            qtmp[i] = qcond_sub[i0:i1].max(axis=0)
        else:
            qtmp[i] = INDEX_MIN
    qcond_cell0 = np.empty((nbas_bvk, nbasp), dtype=qcond.dtype)
    for j, (j0, j1) in enumerate(zip(seg_loc[:nbasp], seg_loc[1:nbasp+1])):
        if j0 != j1:
            qcond_cell0[:,j] = qtmp[:,j0:j1].max(axis=1)
        else:
            qcond_cell0[:,j] = INDEX_MIN
    return qcond_cell0



#!/usr/bin/env python
# Copyright 2014-2019 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''Non-relativistic RKS analytical nuclear gradients'''


import numpy
from pyscf import gto
from pyscf import lib
from pyscf.lib import logger
from pyscf.grad import rhf as rhf_grad
from pyscf.dft import numint, radi, gen_grid, xc_deriv
from pyscf import __config__
import ctypes

libdft = lib.load_library('libdft')


def get_veff(ks_grad, mol=None, dm=None):
    '''
    First order derivative of DFT effective potential matrix (wrt electron coordinates)

    Args:
        ks_grad : grad.uhf.Gradients or grad.uks.Gradients object
    '''
    if mol is None: mol = ks_grad.mol
    if dm is None: dm = ks_grad.base.make_rdm1()
    t0 = (logger.process_clock(), logger.perf_counter())

    mf = ks_grad.base
    ni = mf._numint
    grids, nlcgrids = _initialize_grids(ks_grad)

    mem_now = lib.current_memory()[0]
    max_memory = max(2000, ks_grad.max_memory*.9-mem_now)
    if ks_grad.grid_response:
        exc, vxc = get_vxc_full_response(ni, mol, grids, mf.xc, dm,
                                         max_memory=max_memory,
                                         verbose=ks_grad.verbose)
        if mf.do_nlc():
            if ni.libxc.is_nlc(mf.xc):
                xc = mf.xc
            else:
                xc = mf.nlc
            enlc, vnlc = get_nlc_vxc_full_response(
                ni, mol, nlcgrids, xc, dm,
                max_memory=max_memory, verbose=ks_grad.verbose)
            exc += enlc
            vxc += vnlc
        logger.debug1(ks_grad, 'sum(grids response) %s', exc.sum(axis=0))
    else:
        exc, vxc = get_vxc(ni, mol, grids, mf.xc, dm,
                           max_memory=max_memory, verbose=ks_grad.verbose)
        if mf.do_nlc():
            if ni.libxc.is_nlc(mf.xc):
                xc = mf.xc
            else:
                xc = mf.nlc
            enlc, vnlc = get_nlc_vxc(
                ni, mol, nlcgrids, xc, dm,
                max_memory=max_memory, verbose=ks_grad.verbose)
            vxc += vnlc
    t0 = logger.timer(ks_grad, 'vxc', *t0)

    if not ni.libxc.is_hybrid_xc(mf.xc):
        vj = ks_grad.get_j(mol, dm)
        vxc += vj
    else:
        omega, alpha, hyb = ni.rsh_and_hybrid_coeff(mf.xc, spin=mol.spin)
        vj, vk = ks_grad.get_jk(mol, dm)
        vk *= hyb
        if omega != 0:
            vk += ks_grad.get_k(mol, dm, omega=omega) * (alpha - hyb)
        vxc += vj - vk * .5

    return lib.tag_array(vxc, exc1_grid=exc)

def _initialize_grids(ks_grad):
    mf = ks_grad.base
    if ks_grad.grids is not None:
        grids = ks_grad.grids
    else:
        grids = mf.grids
    if grids.coords is None:
        grids.build(with_non0tab=True)

    nlcgrids = None
    if mf.do_nlc():
        if ks_grad.nlcgrids is not None:
            nlcgrids = ks_grad.nlcgrids
        else:
            nlcgrids = mf.nlcgrids
        if nlcgrids.coords is None:
            nlcgrids.build(with_non0tab=True)
    return grids, nlcgrids


def get_vxc(ni, mol, grids, xc_code, dms, relativity=0, hermi=1,
            max_memory=2000, verbose=None):
    xctype = ni._xc_type(xc_code)
    make_rho, nset, nao = ni._gen_rho_evaluator(mol, dms, hermi, False, grids)
    ao_loc = mol.ao_loc_nr()

    vmat = numpy.zeros((nset,3,nao,nao))
    if xctype == 'LDA':
        ao_deriv = 1
        for ao, mask, weight, coords \
                in ni.block_loop(mol, grids, nao, ao_deriv, max_memory):
            for idm in range(nset):
                rho = make_rho(idm, ao[0], mask, xctype)
                vxc = ni.eval_xc_eff(xc_code, rho, 1, xctype=xctype)[1]
                wv = weight * vxc[0]
                aow = numint._scale_ao(ao[0], wv)
                _d1_dot_(vmat[idm], mol, ao[1:4], aow, mask, ao_loc, True)

    elif xctype == 'GGA':
        ao_deriv = 2
        for ao, mask, weight, coords \
                in ni.block_loop(mol, grids, nao, ao_deriv, max_memory):
            for idm in range(nset):
                rho = make_rho(idm, ao[:4], mask, xctype)
                vxc = ni.eval_xc_eff(xc_code, rho, 1, xctype=xctype)[1]
                wv = weight * vxc
                wv[0] *= .5
                _gga_grad_sum_(vmat[idm], mol, ao, wv, mask, ao_loc)

    elif xctype == 'MGGA':
        ao_deriv = 2
        for ao, mask, weight, coords \
                in ni.block_loop(mol, grids, nao, ao_deriv, max_memory):
            for idm in range(nset):
                rho = make_rho(idm, ao[:10], mask, xctype)
                vxc = ni.eval_xc_eff(xc_code, rho, 1, xctype=xctype)[1]
                wv = weight * vxc
                wv[0] *= .5
                wv[4] *= .5  # for the factor 1/2 in tau
                _gga_grad_sum_(vmat[idm], mol, ao, wv, mask, ao_loc)
                _tau_grad_dot_(vmat[idm], mol, ao, wv[4], mask, ao_loc, True)

    exc = None
    if nset == 1:
        vmat = vmat[0]
    # - sign because nabla_X = -nabla_x
    return exc, -vmat

def get_nlc_vxc(ni, mol, grids, xc_code, dm, relativity=0, hermi=1,
                max_memory=2000, verbose=None):
    make_rho, nset, nao = ni._gen_rho_evaluator(mol, dm, hermi, False, grids)
    assert nset == 1
    ao_loc = mol.ao_loc_nr()

    vmat = numpy.zeros((3,nao,nao))
    nlc_coefs = ni.nlc_coeff(xc_code)
    if len(nlc_coefs) != 1:
        raise NotImplementedError('Additive NLC')
    nlc_pars, fac = nlc_coefs[0]
    ao_deriv = 2
    vvrho = []
    for ao, mask, weight, coords \
            in ni.block_loop(mol, grids, nao, ao_deriv, max_memory):
        vvrho.append(make_rho(0, ao[:4], mask, 'GGA'))
    rho = numpy.hstack(vvrho)

    vxc = numint._vv10nlc(rho, grids.coords, rho, grids.weights,
                          grids.coords, nlc_pars)[1]
    vv_vxc = xc_deriv.transform_vxc(rho, vxc, 'GGA', spin=0)

    p1 = 0
    for ao, mask, weight, coords \
            in ni.block_loop(mol, grids, nao, ao_deriv, max_memory):
        p0, p1 = p1, p1 + weight.size
        wv = vv_vxc[:,p0:p1] * weight
        wv[0] *= .5  # *.5 because vmat + vmat.T at the end
        _gga_grad_sum_(vmat, mol, ao, wv, mask, ao_loc)

    exc = None
    # - sign because nabla_X = -nabla_x
    return exc, -vmat

def _make_dR_dao_w(ao, wv):
    #:aow = numpy.einsum('npi,p->npi', ao[1:4], wv[0])
    aow = [
        numint._scale_ao(ao[1], wv[0]),  # dX nabla_x
        numint._scale_ao(ao[2], wv[0]),  # dX nabla_y
        numint._scale_ao(ao[3], wv[0]),  # dX nabla_z
    ]
    # XX, XY, XZ = 4, 5, 6
    # YX, YY, YZ = 5, 7, 8
    # ZX, ZY, ZZ = 6, 8, 9
    aow[0] += numint._scale_ao(ao[4], wv[1])  # dX nabla_x
    aow[0] += numint._scale_ao(ao[5], wv[2])  # dX nabla_y
    aow[0] += numint._scale_ao(ao[6], wv[3])  # dX nabla_z
    aow[1] += numint._scale_ao(ao[5], wv[1])  # dY nabla_x
    aow[1] += numint._scale_ao(ao[7], wv[2])  # dY nabla_y
    aow[1] += numint._scale_ao(ao[8], wv[3])  # dY nabla_z
    aow[2] += numint._scale_ao(ao[6], wv[1])  # dZ nabla_x
    aow[2] += numint._scale_ao(ao[8], wv[2])  # dZ nabla_y
    aow[2] += numint._scale_ao(ao[9], wv[3])  # dZ nabla_z
    return aow

def _d1_dot_(vmat, mol, ao1, ao2, mask, ao_loc, dR1_on_bra=True):
    shls_slice = (0, mol.nbas)
    if dR1_on_bra:
        vmat[0] += numint._dot_ao_ao(mol, ao1[0], ao2, mask, shls_slice, ao_loc)
        vmat[1] += numint._dot_ao_ao(mol, ao1[1], ao2, mask, shls_slice, ao_loc)
        vmat[2] += numint._dot_ao_ao(mol, ao1[2], ao2, mask, shls_slice, ao_loc)
    else:
        vmat[0] += numint._dot_ao_ao(mol, ao1, ao2[0], mask, shls_slice, ao_loc)
        vmat[1] += numint._dot_ao_ao(mol, ao1, ao2[1], mask, shls_slice, ao_loc)
        vmat[2] += numint._dot_ao_ao(mol, ao1, ao2[2], mask, shls_slice, ao_loc)

def _gga_grad_sum_(vmat, mol, ao, wv, mask, ao_loc):
    #:aow = numpy.einsum('npi,np->pi', ao[:4], wv[:4])
    aow = numint._scale_ao(ao[:4], wv[:4])
    _d1_dot_(vmat, mol, ao[1:4], aow, mask, ao_loc, True)
    aow = _make_dR_dao_w(ao, wv[:4])
    _d1_dot_(vmat, mol, aow, ao[0], mask, ao_loc, True)
    return vmat

# XX, XY, XZ = 4, 5, 6
# YX, YY, YZ = 5, 7, 8
# ZX, ZY, ZZ = 6, 8, 9
def _tau_grad_dot_(vmat, mol, ao, wv, mask, ao_loc, dR1_on_bra=True):
    '''The tau part of MGGA functional'''
    aow = numint._scale_ao(ao[1], wv)
    _d1_dot_(vmat, mol, [ao[4], ao[5], ao[6]], aow, mask, ao_loc, True)
    aow = numint._scale_ao(ao[2], wv, aow)
    _d1_dot_(vmat, mol, [ao[5], ao[7], ao[8]], aow, mask, ao_loc, True)
    aow = numint._scale_ao(ao[3], wv, aow)
    _d1_dot_(vmat, mol, [ao[6], ao[8], ao[9]], aow, mask, ao_loc, True)

def _vv10nlc_grad(rho, coords, vvrho, vvweight, vvcoords, nlc_pars):
    # VV10 gradient term from Vydrov and Van Voorhis 2010 eq. 25-26
    # https://doi.org/10.1063/1.3521275
    thresh=1e-8

    #output
    exc=numpy.zeros((rho[0,:].size,3))

    #outer grid needs threshing
    threshind=rho[0,:]>=thresh
    coords=coords[threshind]
    R=rho[0,:][threshind]
    Gx=rho[1,:][threshind]
    Gy=rho[2,:][threshind]
    Gz=rho[3,:][threshind]
    G=Gx**2.+Gy**2.+Gz**2.

    #inner grid needs threshing
    innerthreshind=vvrho[0,:]>=thresh
    vvcoords=vvcoords[innerthreshind]
    vvweight=vvweight[innerthreshind]
    Rp=vvrho[0,:][innerthreshind]
    RpW=Rp*vvweight
    Gxp=vvrho[1,:][innerthreshind]
    Gyp=vvrho[2,:][innerthreshind]
    Gzp=vvrho[3,:][innerthreshind]
    Gp=Gxp**2.+Gyp**2.+Gzp**2.

    #constants and parameters
    Pi=numpy.pi
    Pi43=4.*Pi/3.
    Bvv, Cvv = nlc_pars
    Kvv=Bvv*1.5*Pi*((9.*Pi)**(-1./6.))
    Beta=((3./(Bvv*Bvv))**(0.75))/32.

    #inner grid
    W0p=Gp/(Rp*Rp)
    W0p=Cvv*W0p*W0p
    W0p=(W0p+Pi43*Rp)**0.5
    Kp=Kvv*(Rp**(1./6.))

    #outer grid
    W0tmp=G/(R**2)
    W0tmp=Cvv*W0tmp*W0tmp
    W0=(W0tmp+Pi43*R)**0.5
    K=Kvv*(R**(1./6.))

    vvcoords = numpy.asarray(vvcoords, order='C')
    coords = numpy.asarray(coords, order='C')
    F = numpy.empty(R.shape +(3,), order='C')
    libdft.VXC_vv10nlc_grad(F.ctypes.data_as(ctypes.c_void_p),
                            vvcoords.ctypes.data_as(ctypes.c_void_p),
                            coords.ctypes.data_as(ctypes.c_void_p),
                            W0p.ctypes.data_as(ctypes.c_void_p),
                            W0.ctypes.data_as(ctypes.c_void_p),
                            K.ctypes.data_as(ctypes.c_void_p),
                            Kp.ctypes.data_as(ctypes.c_void_p),
                            RpW.ctypes.data_as(ctypes.c_void_p),
                            ctypes.c_int(vvcoords.shape[0]),
                            ctypes.c_int(coords.shape[0]))
    #exc is multiplied by Rho later
    exc[threshind] = F
    return exc, Beta

def get_vxc_full_response(ni, mol, grids, xc_code, dms, relativity=0, hermi=1,
                          max_memory=2000, verbose=None):
    '''Full response including the response of the grids'''
    xctype = ni._xc_type(xc_code)
    make_rho, nset, nao = ni._gen_rho_evaluator(mol, dms, hermi, False, grids)
    ao_loc = mol.ao_loc_nr()

    excsum = numpy.zeros((mol.natm,3))
    vmat = numpy.zeros((3,nao,nao))
    if xctype == 'LDA':
        ao_deriv = 1
        vtmp = numpy.empty((3,nao,nao))
        for atm_id, (coords, weight, weight1) in enumerate(grids_response_cc(grids)):
            mask = gen_grid.make_mask(mol, coords)
            ao = ni.eval_ao(mol, coords, deriv=ao_deriv, non0tab=mask,
                            cutoff=grids.cutoff)
            rho = make_rho(0, ao[0], mask, xctype)
            exc, vxc = ni.eval_xc_eff(xc_code, rho, 1, xctype=xctype)[:2]
            wv = weight * vxc[0]

            vtmp = numpy.zeros((3,nao,nao))
            aow = numint._scale_ao(ao[0], wv)
            _d1_dot_(vtmp, mol, ao[1:4], aow, mask, ao_loc, True)
            vmat += vtmp

            # response of weights
            excsum += numpy.einsum('r,r,nxr->nx', exc, rho, weight1)
            # response of grids coordinates
            excsum[atm_id] += numpy.einsum('xij,ji->x', vtmp, dms) * 2
            rho = vxc = aow = None

    elif xctype == 'GGA':
        ao_deriv = 2
        for atm_id, (coords, weight, weight1) in enumerate(grids_response_cc(grids)):
            mask = gen_grid.make_mask(mol, coords)
            ao = ni.eval_ao(mol, coords, deriv=ao_deriv, non0tab=mask,
                            cutoff=grids.cutoff)
            rho = make_rho(0, ao[:4], mask, xctype)
            exc, vxc = ni.eval_xc_eff(xc_code, rho, 1, xctype=xctype)[:2]
            wv = weight * vxc
            wv[0] *= .5

            vtmp = numpy.zeros((3,nao,nao))
            _gga_grad_sum_(vtmp, mol, ao, wv, mask, ao_loc)
            vmat += vtmp

            # response of weights
            excsum += numpy.einsum('r,r,nxr->nx', exc, rho[0], weight1)
            # response of grids coordinates
            excsum[atm_id] += numpy.einsum('xij,ji->x', vtmp, dms) * 2
            rho = vxc = wv = None

    elif xctype == 'MGGA':
        ao_deriv = 2
        for atm_id, (coords, weight, weight1) in enumerate(grids_response_cc(grids)):
            mask = gen_grid.make_mask(mol, coords)
            ao = ni.eval_ao(mol, coords, deriv=ao_deriv, non0tab=mask,
                            cutoff=grids.cutoff)
            rho = make_rho(0, ao[:10], mask, xctype)
            exc, vxc = ni.eval_xc_eff(xc_code, rho, 1, xctype=xctype)[:2]
            wv = weight * vxc
            wv[0] *= .5
            wv[4] *= .5  # for the factor 1/2 in tau

            vtmp = numpy.zeros((3,nao,nao))
            _gga_grad_sum_(vtmp, mol, ao, wv, mask, ao_loc)
            _tau_grad_dot_(vtmp, mol, ao, wv[4], mask, ao_loc, True)
            vmat += vtmp

            # response of weights
            excsum += numpy.einsum('r,r,nxr->nx', exc, rho[0], weight1)
            # response of grids coordinates
            excsum[atm_id] += numpy.einsum('xij,ji->x', vtmp, dms) * 2
            rho = vxc = wv = None

    # - sign because nabla_X = -nabla_x
    return excsum, -vmat

def get_nlc_vxc_full_response(ni, mol, grids, xc_code, dms, relativity=0, hermi=1,
                              max_memory=2000, verbose=None):
    '''Full NLC functional response including the response of the grids'''
    make_rho, nset, nao = ni._gen_rho_evaluator(mol, dms, hermi, False, grids)
    ao_loc = mol.ao_loc_nr()

    excsum = numpy.zeros((mol.natm,3))
    vmat = numpy.zeros((3,nao,nao))
    nlc_coefs = ni.nlc_coeff(xc_code)
    if len(nlc_coefs) != 1:
        raise NotImplementedError('Additive NLC')
    nlc_pars, fac = nlc_coefs[0]
    ao_deriv = 2
    vvrho = []
    vvcoords = []
    vvweights = []
    for atm_id, (coords, weight) in enumerate(grids_noresponse_cc(grids)):
        mask = gen_grid.make_mask(mol, coords)
        ao = ni.eval_ao(mol, coords, deriv=ao_deriv, non0tab=mask,
                        cutoff=grids.cutoff)
        vvrho.append(make_rho(0, ao[:4], mask, 'GGA'))
        vvcoords.append(coords)
        vvweights.append(weight)
    vvcoords_flat = numpy.vstack(vvcoords)
    vvweights_flat = numpy.concatenate(vvweights)
    vvrho_flat = numpy.hstack(vvrho)

    vv_vxc = []
    for atm_id, (coords, weight, weight1) in enumerate(grids_response_cc(grids)):
        rho = vvrho[atm_id]
        mask = gen_grid.make_mask(mol, coords)
        ao = ni.eval_ao(mol, coords, deriv=ao_deriv, non0tab=mask,
                        cutoff=grids.cutoff)

        exc, vxc = numint._vv10nlc(rho, coords, vvrho_flat, vvweights_flat,
                                   vvcoords_flat, nlc_pars)
        vv_vxc = xc_deriv.transform_vxc(rho, vxc, 'GGA', spin=0)
        wv = vv_vxc * weight
        wv[0] *= .5
        vtmp = numpy.zeros((3,nao,nao))
        _gga_grad_sum_(vtmp, mol, ao, wv, mask, ao_loc)
        vmat += vtmp

        vvrho_sub = numpy.hstack(
            [r for i, r in enumerate(vvrho) if i != atm_id])
        vvcoords_sub = numpy.vstack(
            [r for i, r in enumerate(vvcoords) if i != atm_id])
        vvweights_sub = numpy.concatenate(
            [r for i, r in enumerate(vvweights) if i != atm_id])
        egrad, Beta = _vv10nlc_grad(rho, coords, vvrho_sub,
                                    vvweights_sub, vvcoords_sub, nlc_pars)
        # account for factor of 2 in double integration
        exc -= 0.5 * Beta
        # response of weights
        excsum += 2 * numpy.einsum('r,r,nxr->nx', exc, rho[0], weight1)
        # response of grids coordinates
        excsum[atm_id] += 2 * numpy.einsum('xij,ji->x', vtmp, dms)
        excsum[atm_id] += numpy.einsum('r,rx->x', rho[0]*weight, egrad)
    # - sign because nabla_X = -nabla_x
    return excsum, -vmat


# JCP 98, 5612 (1993); DOI:10.1063/1.464906
def grids_response_cc(grids):
    mol = grids.mol
    atom_grids_tab = grids.gen_atomic_grids(mol, grids.atom_grid,
                                            grids.radi_method,
                                            grids.level, grids.prune)
    atm_coords = numpy.asarray(mol.atom_coords() , order='C')
    atm_dist = gto.inter_distance(mol, atm_coords)

    def _radii_adjust(mol, atomic_radii):
        charges = mol.atom_charges()
        if grids.radii_adjust == radi.treutler_atomic_radii_adjust:
            rad = numpy.sqrt(atomic_radii[charges]) + 1e-200
        elif grids.radii_adjust == radi.becke_atomic_radii_adjust:
            rad = atomic_radii[charges] + 1e-200
        else:
            fadjust = lambda i, j, g: g
            gadjust = lambda *args: 1
            return fadjust, gadjust

        rr = rad.reshape(-1,1) * (1./rad)
        a = .25 * (rr.T - rr)
        a[a<-.5] = -.5
        a[a>0.5] = 0.5

        def fadjust(i, j, g):
            return g + a[i,j]*(1-g**2)

        #: d[g + a[i,j]*(1-g**2)] /dg = 1 - 2*a[i,j]*g
        def gadjust(i, j, g):
            return 1 - 2*a[i,j]*g
        return fadjust, gadjust

    fadjust, gadjust = _radii_adjust(mol, grids.atomic_radii)

    def gen_grid_partition(coords, atom_id):
        ngrids = coords.shape[0]
        grid_dist = []
        grid_norm_vec = []
        for ia in range(mol.natm):
            v = (atm_coords[ia] - coords).T
            normv = numpy.linalg.norm(v,axis=0) + 1e-200
            v /= normv
            grid_dist.append(normv)
            grid_norm_vec.append(v)

        def get_du(ia, ib):  # JCP 98, 5612 (1993); (B10)
            uab = atm_coords[ia] - atm_coords[ib]
            duab = 1./atm_dist[ia,ib] * grid_norm_vec[ia]
            duab-= uab[:,None]/atm_dist[ia,ib]**3 * (grid_dist[ia]-grid_dist[ib])
            return duab

        pbecke = numpy.ones((mol.natm,ngrids))
        dpbecke = numpy.zeros((mol.natm,mol.natm,3,ngrids))
        for ia in range(mol.natm):
            for ib in range(ia):
                g = 1/atm_dist[ia,ib] * (grid_dist[ia]-grid_dist[ib])
                p0 = fadjust(ia, ib, g)
                p1 = (3 - p0**2) * p0 * .5
                p2 = (3 - p1**2) * p1 * .5
                p3 = (3 - p2**2) * p2 * .5
                t_uab = 27./16 * (1-p2**2) * (1-p1**2) * (1-p0**2) * gadjust(ia, ib, g)

                s_uab = .5 * (1 - p3 + 1e-200)
                s_uba = .5 * (1 + p3 + 1e-200)
                pbecke[ia] *= s_uab
                pbecke[ib] *= s_uba
                pt_uab =-t_uab / s_uab
                pt_uba = t_uab / s_uba

# * When grid is on atom ia/ib, ua/ub == 0, d_uba/d_uab may have huge error
#   How to remove this error?
                duab = get_du(ia, ib)
                duba = get_du(ib, ia)
                if ia == atom_id:
                    dpbecke[ia,ia] += pt_uab * duba
                    dpbecke[ia,ib] += pt_uba * duba
                else:
                    dpbecke[ia,ia] += pt_uab * duab
                    dpbecke[ia,ib] += pt_uba * duab

                if ib == atom_id:
                    dpbecke[ib,ib] -= pt_uba * duab
                    dpbecke[ib,ia] -= pt_uab * duab
                else:
                    dpbecke[ib,ib] -= pt_uba * duba
                    dpbecke[ib,ia] -= pt_uab * duba

# * JCP 98, 5612 (1993); (B8) (B10) miss many terms
                if ia != atom_id and ib != atom_id:
                    ua_ub = grid_norm_vec[ia] - grid_norm_vec[ib]
                    ua_ub /= atm_dist[ia,ib]
                    dpbecke[atom_id,ia] -= pt_uab * ua_ub
                    dpbecke[atom_id,ib] -= pt_uba * ua_ub

        for ia in range(mol.natm):
            dpbecke[:,ia] *= pbecke[ia]

        return pbecke, dpbecke

    natm = mol.natm
    for ia in range(natm):
        coords, vol = atom_grids_tab[mol.atom_symbol(ia)]
        coords = coords + atm_coords[ia]
        pbecke, dpbecke = gen_grid_partition(coords, ia)
        z = 1./pbecke.sum(axis=0)
        w1 = dpbecke[:,ia] * z
        w1 -= pbecke[ia] * z**2 * dpbecke.sum(axis=1)
        w1 *= vol
        w0 = vol * pbecke[ia] * z
        yield coords, w0, w1


def grids_noresponse_cc(grids):
    # same as above but without the response, for nlc grids response routine
    assert grids.becke_scheme == gen_grid.original_becke
    mol = grids.mol
    atom_grids_tab = grids.gen_atomic_grids(mol, grids.atom_grid,
                                            grids.radi_method,
                                            grids.level, grids.prune)
    coords_all, weights_all = gen_grid.get_partition(mol, atom_grids_tab,
                                                     grids.radii_adjust,
                                                     grids.atomic_radii,
                                                     grids.becke_scheme,
                                                     concat=False)
    natm = mol.natm
    for ia in range(natm):
        yield coords_all[ia], weights_all[ia]


class Gradients(rhf_grad.Gradients):

    # This parameter has no effects for HF gradients. Add this attribute so that
    # the kernel function can be reused in the DFT gradients code.
    grid_response = getattr(__config__, 'grad_rks_Gradients_grid_response', False)

    _keys = {'grid_response', 'grids', 'nlcgrids'}

    def __init__(self, mf):
        rhf_grad.Gradients.__init__(self, mf)
        self.grids = None
        self.nlcgrids = None

    def dump_flags(self, verbose=None):
        rhf_grad.Gradients.dump_flags(self, verbose)
        logger.info(self, 'grid_response = %s', self.grid_response)
        #if callable(self.base.grids.prune):
        #    logger.info(self, 'Grid pruning %s may affect DFT gradients accuracy.'
        #                'Call mf.grids.run(prune=False) to mute grid pruning',
        #                self.base.grids.prune)
        return self

    get_veff = get_veff

    def extra_force(self, atom_id, envs):
        '''Hook for extra contributions in analytical gradients.

        Contributions like the response of auxiliary basis in density fitting
        method, the grid response in DFT numerical integration can be put in
        this function.
        '''
        if self.grid_response:
            vhf = envs['vhf']
            log = envs['log']
            log.debug('grids response for atom %d %s',
                      atom_id, vhf.exc1_grid[atom_id])
            return vhf.exc1_grid[atom_id]
        else:
            return 0

Grad = Gradients

from pyscf import dft
dft.rks.RKS.Gradients = dft.rks_symm.RKS.Gradients = lib.class_as_method(Gradients)
dft.roks.ROKS.Gradients = lib.invalid_method('Gradients')



#!/usr/bin/env python
# Copyright 2021-2024 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Xing Zhang <zhangxing.nju@gmail.com>
#

import ctypes
import numpy as np
from pyscf import __config__
from pyscf import lib
from pyscf.lib import logger

libdft = lib.load_library('libdft')

def gradient_gs(f_gs, Gv):
    r'''Compute the G-space components of :math:`\nabla f(r)`
    given :math:`f(G)` and :math:`G`.

    This is equivalent to einsum('np,px->nxp', f_gs, 1j*Gv)
    with multithreading.
    '''
    ng, dim = Gv.shape
    assert dim == 3
    Gv = np.asarray(Gv, order='C', dtype=np.double)
    f_gs = np.asarray(f_gs.reshape(-1,ng), order='C', dtype=np.complex128)
    n = f_gs.shape[0]
    out = np.empty((n,dim,ng), dtype=np.complex128)
    libdft.gradient_gs(
        out.ctypes.data_as(ctypes.c_void_p),
        f_gs.ctypes.data_as(ctypes.c_void_p),
        Gv.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_int(n), ctypes.c_size_t(ng)
    )
    return out

def get_gga_vrho_gs(v, v1, Gv, weight, ngrid, fac=2.):
    '''Update v inplace
    v -= fac * 1j * np.einsum('px,xp->p', Gv, v1)
    v *= weight
    '''
    v = np.asarray(v, order='C', dtype=np.complex128)
    v1 = np.asarray(v1, order='C', dtype=np.complex128)
    Gv = np.asarray(Gv, order='C', dtype=np.double)

    libdft.get_gga_vrho_gs(
        v.ctypes.data_as(ctypes.c_void_p),
        v1.ctypes.data_as(ctypes.c_void_p),
        Gv.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_double(fac),
        ctypes.c_double(weight),
        ctypes.c_int(ngrid)
    )
    return v

def pp_loc_part1_gs(coulG, Gv, G2, G0idx, Z, coords, rloc):
    libpbc = lib.load_library('libpbc')
    coulG = np.asarray(coulG, order='C', dtype=np.double)
    Gv = np.asarray(Gv, order='C', dtype=np.double)
    G2 = np.asarray(G2, order='C', dtype=np.double)
    ngrid = len(G2)

    coords = np.asarray(coords, order='C', dtype=np.double)
    Z = np.asarray(Z, order='C', dtype=np.double)
    rloc = np.asarray(rloc, order='C', dtype=np.double)
    natm = len(Z)

    out = np.empty(ngrid, dtype=np.complex128)
    libpbc.pp_loc_part1_gs(
        out.ctypes.data_as(ctypes.c_void_p),
        coulG.ctypes.data_as(ctypes.c_void_p),
        Gv.ctypes.data_as(ctypes.c_void_p),
        G2.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_int(G0idx),
        ctypes.c_int(ngrid),
        Z.ctypes.data_as(ctypes.c_void_p),
        coords.ctypes.data_as(ctypes.c_void_p),
        rloc.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_int(natm))
    return out

def build_core_density(
    atm,
    bas,
    env,
    mesh,
    dimension,
    a,
    b,
    max_radius,
):
    if abs(a - np.diag(a.diagonal())).max() < 1e-12:
        lattice_type = '_orth'
        orth = True
    else:
        lattice_type = '_nonorth'
        orth = False
    fn_name = 'make_rho_lda' + lattice_type

    atm = np.asarray(atm, order='C', dtype=np.int32)
    bas = np.asarray(bas, order='C', dtype=np.int32)
    env = np.asarray(env, order='C', dtype=np.double)
    a = np.asarray(a, order='C', dtype=np.double)
    b = np.asarray(b, order='C', dtype=np.double)
    mesh = np.asarray(mesh, order='C', dtype=np.int32)
    rho_core = np.zeros((np.prod(mesh),))

    libdft.build_core_density(
        getattr(libdft, fn_name),
        rho_core.ctypes.data_as(ctypes.c_void_p),
        atm.ctypes.data_as(ctypes.c_void_p),
        bas.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_int(len(bas)),
        env.ctypes.data_as(ctypes.c_void_p),
        mesh.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_int(dimension),
        a.ctypes.data_as(ctypes.c_void_p),
        b.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_double(max_radius),
        ctypes.c_bool(orth)
    )
    return rho_core


def int_gauss_charge_v_rs(
    v_rs,
    comp,
    atm,
    bas,
    env,
    mesh,
    dimension,
    a,
    b,
    max_radius,
):
    if abs(a - np.diag(a.diagonal())).max() < 1e-12:
        lattice_type = '_orth'
        orth = True
    else:
        lattice_type = '_nonorth'
        orth = False

    fn_name = 'eval_mat_lda' + lattice_type
    if comp == 3:
        fn_name += '_ip1'
    elif comp != 1:
        raise NotImplementedError

    out = np.zeros((len(atm), comp), order='C', dtype=np.double)
    v_rs = np.asarray(v_rs, order='C', dtype=np.double)
    atm = np.asarray(atm, order='C', dtype=np.int32)
    bas = np.asarray(bas, order='C', dtype=np.int32)
    env = np.asarray(env, order='C', dtype=np.double)
    mesh = np.asarray(mesh, order='C', dtype=np.int32)
    a = np.asarray(a, order='C', dtype=np.double)
    b = np.asarray(b, order='C', dtype=np.double)

    libdft.int_gauss_charge_v_rs(
        getattr(libdft, fn_name),
        out.ctypes.data_as(ctypes.c_void_p),
        v_rs.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_int(comp),
        atm.ctypes.data_as(ctypes.c_void_p),
        bas.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_int(len(bas)),
        env.ctypes.data_as(ctypes.c_void_p),
        mesh.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_int(dimension),
        a.ctypes.data_as(ctypes.c_void_p),
        b.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_double(max_radius),
        ctypes.c_bool(orth)
    )
    return out


def grid_collocate(
    xctype,
    dm,
    task_list,
    hermi,
    shls_slice,
    ao_loc0,
    ao_loc1,
    dimension,
    a,
    b,
    ish_atm,
    ish_bas,
    ish_env,
    jsh_atm,
    jsh_bas,
    jsh_env,
    cart,
):
    if abs(a-np.diag(a.diagonal())).max() < 1e-12:
        lattice_type = '_orth'
        orth = True
    else:
        lattice_type = '_nonorth'
        orth = False

    if xctype.upper() == 'LDA':
        comp = 1
    else:
        raise NotImplementedError

    fn_name = 'make_rho_' + xctype.lower() + lattice_type

    rs_rho = RS_Grid(task_list.gridlevel_info, comp)
    dm = np.asarray(dm, order='C', dtype=np.double)
    i0, i1, j0, j1 = shls_slice
    ao_loc0 = np.asarray(ao_loc0, order='C', dtype=np.int32)
    ao_loc1 = np.asarray(ao_loc1, order='C', dtype=np.int32)
    Ls = np.asarray(task_list.Ls, order='C', dtype=np.double)
    a = np.asarray(a, order='C', dtype=np.double)
    b = np.asarray(b, order='C', dtype=np.double)
    ish_atm = np.asarray(ish_atm, order='C', dtype=np.int32)
    ish_bas = np.asarray(ish_bas, order='C', dtype=np.int32)
    ish_env = np.asarray(ish_env, order='C', dtype=np.double)
    jsh_atm = np.asarray(jsh_atm, order='C', dtype=np.int32)
    jsh_bas = np.asarray(jsh_bas, order='C', dtype=np.int32)
    jsh_env = np.asarray(jsh_env, order='C', dtype=np.double)

    libdft.grid_collocate_drv(
        getattr(libdft, fn_name),
        ctypes.byref(rs_rho._this),
        dm.ctypes.data_as(ctypes.c_void_p),
        ctypes.byref(task_list._this),
        ctypes.c_int(comp),
        ctypes.c_int(hermi),
        (ctypes.c_int*4)(i0, i1, j0, j1),
        ao_loc0.ctypes.data_as(ctypes.c_void_p),
        ao_loc1.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_int(dimension),
        Ls.ctypes.data_as(ctypes.c_void_p),
        a.ctypes.data_as(ctypes.c_void_p),
        b.ctypes.data_as(ctypes.c_void_p),
        ish_atm.ctypes.data_as(ctypes.c_void_p),
        ish_bas.ctypes.data_as(ctypes.c_void_p),
        ish_env.ctypes.data_as(ctypes.c_void_p),
        jsh_atm.ctypes.data_as(ctypes.c_void_p),
        jsh_bas.ctypes.data_as(ctypes.c_void_p),
        jsh_env.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_int(cart),
        ctypes.c_bool(orth)
    )
    return rs_rho

def grid_integrate(
    xctype,
    wv,
    task_list,
    comp,
    hermi,
    grid_level,
    shls_slice,
    ao_loc0,
    ao_loc1,
    dimension,
    a,
    b,
    ish_atm,
    ish_bas,
    ish_env,
    jsh_atm,
    jsh_bas,
    jsh_env,
    cart,
):
    if abs(a-np.diag(a.diagonal())).max() < 1e-12:
        lattice_type = '_orth'
        orth = True
    else:
        lattice_type = '_nonorth'
        orth = False

    fn_name = 'eval_mat_' + xctype.lower() + lattice_type
    if comp == 3:
        fn_name += '_ip1'
    elif comp != 1:
        raise NotImplementedError

    i0, i1, j0, j1 = shls_slice
    ao_loc0 = np.asarray(ao_loc0, order='C', dtype=np.int32)
    ao_loc1 = np.asarray(ao_loc1, order='C', dtype=np.int32)
    naoi = ao_loc0[i1] - ao_loc0[i0]
    naoj = ao_loc1[j1] - ao_loc1[j0]
    if comp == 1:
        mat = np.zeros((naoi, naoj))
    else:
        mat = np.zeros((comp, naoi, naoj))

    wv = np.asarray(wv, order='C', dtype=np.double)
    Ls = np.asarray(task_list.Ls, order='C', dtype=np.double)
    a = np.asarray(a, order='C', dtype=np.double)
    b = np.asarray(b, order='C', dtype=np.double)
    ish_atm = np.asarray(ish_atm, order='C', dtype=np.int32)
    ish_bas = np.asarray(ish_bas, order='C', dtype=np.int32)
    ish_env = np.asarray(ish_env, order='C', dtype=np.double)
    jsh_atm = np.asarray(jsh_atm, order='C', dtype=np.int32)
    jsh_bas = np.asarray(jsh_bas, order='C', dtype=np.int32)
    jsh_env = np.asarray(jsh_env, order='C', dtype=np.double)

    libdft.grid_integrate_drv(
        getattr(libdft, fn_name),
        mat.ctypes.data_as(ctypes.c_void_p),
        wv.ctypes.data_as(ctypes.c_void_p),
        ctypes.byref(task_list._this),
        ctypes.c_int(comp),
        ctypes.c_int(hermi),
        ctypes.c_int(grid_level),
        (ctypes.c_int*4)(i0, i1, j0, j1),
        ao_loc0.ctypes.data_as(ctypes.c_void_p),
        ao_loc1.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_int(dimension),
        Ls.ctypes.data_as(ctypes.c_void_p),
        a.ctypes.data_as(ctypes.c_void_p),
        b.ctypes.data_as(ctypes.c_void_p),
        ish_atm.ctypes.data_as(ctypes.c_void_p),
        ish_bas.ctypes.data_as(ctypes.c_void_p),
        ish_env.ctypes.data_as(ctypes.c_void_p),
        jsh_atm.ctypes.data_as(ctypes.c_void_p),
        jsh_bas.ctypes.data_as(ctypes.c_void_p),
        jsh_env.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_int(cart),
        ctypes.c_bool(orth)
    )
    return mat


class _CGridLevel_Info(ctypes.Structure):
    '''C structure for `GridLevel_Info`.
    '''
    _fields_ = [("nlevels", ctypes.c_int), # number of grid levels
                ("rel_cutoff", ctypes.c_double),
                ("cutoff", ctypes.POINTER(ctypes.c_double)),
                ("mesh", ctypes.POINTER(ctypes.c_int))]


class GridLevel_Info:
    '''Information of multigrids.
    '''
    def __init__(self, cutoff, rel_cutoff, mesh):
        self._this = ctypes.POINTER(_CGridLevel_Info)()

        if cutoff[0] < 1e-15:
            cutoff = cutoff[1:]

        self.cutoff = np.asarray(cutoff, order='C', dtype=np.double)
        self.mesh = np.asarray(np.asarray(mesh).reshape(-1,3), order='C', dtype=np.int32)
        self.rel_cutoff = rel_cutoff
        self.nlevels = len(cutoff)

        libdft.init_gridlevel_info(
            ctypes.byref(self._this),
            self.cutoff.ctypes.data_as(ctypes.c_void_p),
            self.mesh.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(self.nlevels),
            ctypes.c_double(self.rel_cutoff)
        )

    def __del__(self):
        try:
            libdft.del_gridlevel_info(ctypes.byref(self._this))
        except AttributeError:
            pass


class _CRS_Grid(ctypes.Structure):
    '''C structure for `RS_Grid`.
    '''
    _fields_ = [("nlevels", ctypes.c_int),
                ("gridlevel_info", ctypes.POINTER(_CGridLevel_Info)),
                ("comp", ctypes.c_int),
                # data is list of 1d arrays
                ("data", ctypes.POINTER(ctypes.POINTER(ctypes.c_double)))]


class RS_Grid:
    '''Density values on real-space multigrids.
    '''
    def __init__(self, gridlevel_info, comp):
        self._this = ctypes.POINTER(_CRS_Grid)()
        self.gridlevel_info = gridlevel_info
        self.nlevels = gridlevel_info.nlevels

        libdft.init_rs_grid(
            ctypes.byref(self._this),
            ctypes.byref(gridlevel_info._this),
            ctypes.c_int(comp)
        )

    def __del__(self):
        try:
            libdft.del_rs_grid(ctypes.byref(self._this))
        except AttributeError:
            pass

    def __getitem__(self, i):
        '''Get the i-th level density as a numpy array.
        '''
        ngrids = np.prod(self.gridlevel_info.mesh[i])
        return np.ctypeslib.as_array(self._this.contents.data[i], shape=(ngrids,))


class _CPGFPair(ctypes.Structure):
    '''C structure for a pair of primitive Gaussian functions.
    '''
    _fields_ = [("ish", ctypes.c_int),
                ("ipgf", ctypes.c_int),
                ("jsh", ctypes.c_int),
                ("jpgf", ctypes.c_int),
                ("iL", ctypes.c_int),
                ("radius", ctypes.c_double)]


class _CTask(ctypes.Structure):
    '''C structure for a single task.
    '''
    _fields_ = [("buf_size", ctypes.c_size_t),
                ("ntasks", ctypes.c_size_t),
                ("pgfpairs", ctypes.POINTER(ctypes.POINTER(_CPGFPair))),
                ("radius", ctypes.c_double)]


class _CTaskList(ctypes.Structure):
    '''C structure for a task list.
    '''
    _fields_ = [("nlevels", ctypes.c_int),
                ("hermi", ctypes.c_int),
                ("gridlevel_info", ctypes.POINTER(_CGridLevel_Info)),
                ("tasks", ctypes.POINTER(ctypes.POINTER(_CTask)))]


class TaskList:
    '''Task list for multigrid DFT calculations.
    '''
    grid_level_method = getattr(__config__, "pbc_dft_multigrid_grid_level_method", "pyscf")

    def __init__(self, cell, gridlevel_info,
                 cell1=None, Ls=None, hermi=0, precision=None):
        '''Build a task list.

        Arguments:
            cell : :class:`pbc.gto.cell.Cell`
                The :class:`Cell` instance for the bra basis functions.
            gridlevel_info : :class:`GridLevel_Info`
                Information of the multiple grids.
            cell1 : :class:`pbc.gto.cell.Cell`, optional
                The :class:`Cell` instance for the ket basis functions.
                If not given, both bra and ket basis functions come from `cell`.
            Ls : (*,3) array, optional
                The cartesian coordinates of the periodic images.
                Default is calculated by :func:`cell.get_lattice_Ls`.
            hermi : int, optional
                If :math:`hermi=1`, the task list is built only for
                the upper triangle of the matrix. Default is 0.
            precision : float, optional
                The integral precision. Default is :attr:`cell.precision`.
        '''
        from pyscf.pbc.gto import (
            build_neighbor_list_for_shlpairs,
            free_neighbor_list
        )

        self._this = ctypes.POINTER(_CTaskList)()
        self.gridlevel_info = gridlevel_info
        self.nlevels = gridlevel_info.nlevels

        if cell1 is None:
            cell1 = cell

        if precision is None:
            precision = cell.precision

        if hermi == 1 and cell1 is not cell:
            logger.warn(cell,
                        "Set hermi=0 because cell and cell1 are not the same.")
            hermi = 0
        self.hermi = hermi

        ish_atm = np.asarray(cell._atm, order='C', dtype=np.int32)
        ish_bas = np.asarray(cell._bas, order='C', dtype=np.int32)
        ish_env = np.asarray(cell._env, order='C', dtype=np.double)
        nish = len(ish_bas)
        ish_rcut, ipgf_rcut = cell.rcut_by_shells(precision=precision,
                                                  return_pgf_radius=True)
        assert nish == len(ish_rcut)
        ptr_ipgf_rcut = lib.ndarray_pointer_2d(ipgf_rcut)

        if cell1 is cell:
            jsh_atm = ish_atm
            jsh_bas = ish_bas
            jsh_env = ish_env
            jsh_rcut = ish_rcut
            jpgf_rcut = ipgf_rcut
            ptr_jpgf_rcut = ptr_ipgf_rcut
        else:
            jsh_atm = np.asarray(cell1._atm, order='C', dtype=np.int32)
            jsh_bas = np.asarray(cell1._bas, order='C', dtype=np.int32)
            jsh_env = np.asarray(cell1._env, order='C', dtype=np.double)
            jsh_rcut, jpgf_rcut = cell1.rcut_by_shells(precision=precision,
                                                       return_pgf_radius=True)
            ptr_jpgf_rcut = lib.ndarray_pointer_2d(jpgf_rcut)
        njsh = len(jsh_bas)
        assert njsh == len(jsh_rcut)

        if Ls is None:
            # The default cell.rcut might be insufficient to include all
            # significant pairs
            rcut = ish_rcut.max(initial=0) + jsh_rcut.max(initial=0)
            Ls = cell.get_lattice_Ls(rcut=rcut)
        # The lattice sum vectors must be consistent with the neighbor_list in
        # the backend functions
        self.Ls = Ls
        Ls = np.asarray(Ls, order='C', dtype=np.double)

        nl = build_neighbor_list_for_shlpairs(cell, cell1, Ls=Ls,
                                              ish_rcut=ish_rcut, jsh_rcut=jsh_rcut,
                                              hermi=hermi)

        if self.grid_level_method.lower() == "cp2k":
            fn_name = "get_grid_level_cp2k"
        else:
            fn_name = "get_grid_level"

        libdft.build_task_list(
            ctypes.byref(self._this),
            ctypes.byref(nl),
            ctypes.byref(gridlevel_info._this),
            getattr(libdft, fn_name),
            ish_atm.ctypes.data_as(ctypes.c_void_p),
            ish_bas.ctypes.data_as(ctypes.c_void_p),
            ish_env.ctypes.data_as(ctypes.c_void_p),
            ish_rcut.ctypes.data_as(ctypes.c_void_p),
            ptr_ipgf_rcut.ctypes,
            jsh_atm.ctypes.data_as(ctypes.c_void_p),
            jsh_bas.ctypes.data_as(ctypes.c_void_p),
            jsh_env.ctypes.data_as(ctypes.c_void_p),
            jsh_rcut.ctypes.data_as(ctypes.c_void_p),
            ptr_jpgf_rcut.ctypes,
            ctypes.c_int(nish), ctypes.c_int(njsh),
            Ls.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_double(precision),
            ctypes.c_int(hermi)
        )

    @property
    def ntasks(self):
        return [self._this.contents.tasks[i].contents.ntasks for i in range(self.nlevels)]

    def __del__(self):
        try:
            libdft.del_task_list(ctypes.byref(self._this))
        except AttributeError:
            pass




#!/usr/bin/env python
# Copyright 2014-2019 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Authors: Artem Pulkin, pyscf authors

from pyscf.pbc.lib.kpts_helper import VectorSplitter, VectorComposer
from pyscf.pbc.mp.kmp2 import padding_k_idx
from pyscf.pbc.cc import kccsd_rhf

import numpy as np


def iter_12(cc_or_eom, k):
    """Iterates over EA index slices."""
    if isinstance(cc_or_eom, kccsd_rhf.RCCSD):
        cc = cc_or_eom
    else:
        cc = cc_or_eom._cc
    o, v = padding_k_idx(cc, kind="split")
    kconserv = cc.khelper.kconserv

    yield (v[k],)

    for ki in range(cc.nkpts):
        for ka in range(cc.nkpts):
            kb = kconserv[k, ka, ki]
            yield (ki,), (ka,), o[ki], v[ka], v[kb]


def amplitudes_to_vector(cc_or_eom, t1, t2, kshift=0, kconserv=None):
    """EA amplitudes to vector."""
    itr = iter_12(cc_or_eom, kshift)
    t1, t2 = np.asarray(t1), np.asarray(t2)

    vc = VectorComposer(t1.dtype)
    vc.put(t1[np.ix_(*next(itr))])
    for slc in itr:
        vc.put(t2[np.ix_(*slc)])
    return vc.flush()


def vector_to_amplitudes(cc_or_eom, vec, kshift=0):
    """EA vector to apmplitudes."""
    expected_vs = vector_size(cc_or_eom, kshift)
    if expected_vs != len(vec):
        raise ValueError("The size of the vector passed {:d} should be exactly {:d}".format(len(vec), expected_vs))

    itr = iter_12(cc_or_eom, kshift)
    nocc = cc_or_eom.nocc
    nmo = cc_or_eom.nmo
    nkpts = cc_or_eom.nkpts
    nvirt = nmo - nocc

    vs = VectorSplitter(vec)
    r1 = vs.get(nvirt, slc=next(itr))
    r2 = np.zeros((nkpts, nkpts, nocc, nvirt, nvirt), vec.dtype)
    for slc in itr:
        vs.get(r2, slc=slc)
    return r1, r2


def vector_size(cc_or_eom, kshift=0):
    """The total number of elements in EA vector."""
    size = 0
    for slc in iter_12(cc_or_eom, kshift):
        size += np.prod(tuple(len(i) for i in slc))
    return size



# Copyright 2014-2018 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#from . import berny_solver as berny

from .addons import as_pyscf_method

def optimize(method, *args, **kwargs):
    try:
        from . import geometric_solver as geom
    except ImportError as e1:
        try:
            from . import berny_solver as geom
        except ImportError:
            raise e1
    return geom.optimize(method, *args, **kwargs)



#!/usr/bin/env python
# Copyright 2014-2021 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
FCI solver for Singlet state

Different FCI solvers are implemented to support different type of symmetry.
                    Symmetry
File                Point group   Spin singlet   Real hermitian*    Alpha/beta degeneracy
direct_spin0_symm   Yes           Yes            Yes                Yes
direct_spin1_symm   Yes           No             Yes                Yes
direct_spin0        No            Yes            Yes                Yes
direct_spin1        No            No             Yes                Yes
direct_uhf          No            No             Yes                No
direct_nosym        No            No             No**               Yes

*  Real hermitian Hamiltonian implies (ij|kl) = (ji|kl) = (ij|lk) = (ji|lk)
** Hamiltonian is real but not hermitian, (ij|kl) != (ji|kl) ...

direct_spin0 solver is specified for singlet state. However, calling this
solver sometimes ends up with the error "State not singlet x.xxxxxxe-06" due
to numerical issues. Calling direct_spin1 for singlet state is slightly
slower but more robust than direct_spin0 especially when combining to energy
penalty method (:func:`fix_spin_`)
'''

import ctypes
import numpy
from pyscf import lib
from pyscf import ao2mo
from pyscf.lib import logger
from pyscf.fci import cistring
from pyscf.fci import rdm
from pyscf.fci import direct_spin1
from pyscf.fci.spin_op import contract_ss

libfci = direct_spin1.libfci

@lib.with_doc(direct_spin1.contract_1e.__doc__)
def contract_1e(f1e, fcivec, norb, nelec, link_index=None):
    fcivec = numpy.asarray(fcivec, order='C')
    link_index = direct_spin1._unpack(norb, nelec, link_index)
    if not isinstance(link_index, numpy.ndarray):
        # Handle computability. link_index should be (nparray, nparray)
        link_index = link_index[0]
    na, nlink = link_index.shape[:2]
    assert fcivec.size == na**2
    assert fcivec.dtype == f1e.dtype == numpy.float64
    ci1 = numpy.empty_like(fcivec)
    f1e_tril = lib.pack_tril(f1e)
    libfci.FCIcontract_1e_spin0(f1e_tril.ctypes.data_as(ctypes.c_void_p),
                                fcivec.ctypes.data_as(ctypes.c_void_p),
                                ci1.ctypes.data_as(ctypes.c_void_p),
                                ctypes.c_int(norb), ctypes.c_int(na),
                                ctypes.c_int(nlink),
                                link_index.ctypes.data_as(ctypes.c_void_p))
# no *.5 because FCIcontract_2e_spin0 only compute half of the contraction
    ci1 = lib.transpose_sum(ci1, inplace=True).reshape(fcivec.shape)
    return ci1.view(direct_spin1.FCIvector)

# Note eri is NOT the 2e hamiltonian matrix, the 2e hamiltonian is
# h2e = eri_{pq,rs} p^+ q r^+ s
#     = (pq|rs) p^+ r^+ s q - (pq|rs) \delta_{qr} p^+ s
# so eri is defined as
#       eri_{pq,rs} = (pq|rs) - (1/Nelec) \sum_q (pq|qs)
# to restore the symmetry between pq and rs,
#       eri_{pq,rs} = (pq|rs) - (.5/Nelec) [\sum_q (pq|qs) + \sum_p (pq|rp)]
# Please refer to the treatment in direct_spin1.absorb_h1e
# the input fcivec should be symmetrized
@lib.with_doc(direct_spin1.contract_2e.__doc__)
def contract_2e(eri, fcivec, norb, nelec, link_index=None):
    fcivec = numpy.asarray(fcivec, order='C')
    eri = ao2mo.restore(4, eri, norb)
    lib.transpose_sum(eri, inplace=True)
    eri *= .5
    link_index = direct_spin1._unpack(norb, nelec, link_index)
    if not isinstance(link_index, numpy.ndarray):
        # Handle computability. link_index should be (nparray, nparray)
        link_index = link_index[0]
    na, nlink = link_index.shape[:2]
    assert fcivec.size == na**2
    assert fcivec.dtype == eri.dtype == numpy.float64
    ci1 = numpy.empty((na,na))

    libfci.FCIcontract_2e_spin0(eri.ctypes.data_as(ctypes.c_void_p),
                                fcivec.ctypes.data_as(ctypes.c_void_p),
                                ci1.ctypes.data_as(ctypes.c_void_p),
                                ctypes.c_int(norb), ctypes.c_int(na),
                                ctypes.c_int(nlink),
                                link_index.ctypes.data_as(ctypes.c_void_p))
# no *.5 because FCIcontract_2e_spin0 only compute half of the contraction
    ci1 = lib.transpose_sum(ci1, inplace=True).reshape(fcivec.shape)
    return ci1.view(direct_spin1.FCIvector)

absorb_h1e = direct_spin1.absorb_h1e

@lib.with_doc(direct_spin1.make_hdiag.__doc__)
def make_hdiag(h1e, eri, norb, nelec, compress=False):
    hdiag = direct_spin1.make_hdiag(h1e, eri, norb, nelec)
    na = int(numpy.sqrt(hdiag.size))
# symmetrize hdiag to reduce numerical error
    hdiag = lib.transpose_sum(hdiag.reshape(na,na), inplace=True) * .5
    return hdiag.ravel()

pspace = direct_spin1.pspace

# be careful with single determinant initial guess. It may lead to the
# eigvalue of first davidson iter being equal to hdiag
def kernel(h1e, eri, norb, nelec, ci0=None, level_shift=1e-3, tol=1e-10,
           lindep=1e-14, max_cycle=50, max_space=12, nroots=1,
           davidson_only=False, pspace_size=400, orbsym=None, wfnsym=None,
           ecore=0, **kwargs):
    e, c = direct_spin1._kfactory(FCISolver, h1e, eri, norb, nelec, ci0, level_shift,
                                  tol, lindep, max_cycle, max_space, nroots,
                                  davidson_only, pspace_size, ecore=ecore, **kwargs)
    return e, c

# dm[p,q] = <|q^+ p|>
@lib.with_doc(direct_spin1.make_rdm1.__doc__)
def make_rdm1(fcivec, norb, nelec, link_index=None):
    rdm1 = rdm.make_rdm1('FCImake_rdm1a', fcivec, fcivec,
                         norb, nelec, link_index)
    return rdm1 * 2

# alpha and beta 1pdm
@lib.with_doc(direct_spin1.make_rdm1s.__doc__)
def make_rdm1s(fcivec, norb, nelec, link_index=None):
    rdm1 = rdm.make_rdm1('FCImake_rdm1a', fcivec, fcivec,
                         norb, nelec, link_index)
    return rdm1, rdm1

# Chemist notation
@lib.with_doc(direct_spin1.make_rdm12.__doc__)
def make_rdm12(fcivec, norb, nelec, link_index=None, reorder=True):
    #dm1, dm2 = rdm.make_rdm12('FCIrdm12kern_spin0', fcivec, fcivec,
    #                          norb, nelec, link_index, 1)

    # NOT use FCIrdm12kern_spin0 because for small system, the kernel may call
    # direct diagonalization, which may not fulfil  fcivec = fcivet.T
    dm1, dm2 = rdm.make_rdm12('FCIrdm12kern_sf', fcivec, fcivec,
                              norb, nelec, link_index, 1)
    if reorder:
        dm1, dm2 = rdm.reorder_rdm(dm1, dm2, True)
    return dm1, dm2

# dm[p,q] = <I|q^+ p|J>
@lib.with_doc(direct_spin1.trans_rdm1s.__doc__)
def trans_rdm1s(cibra, ciket, norb, nelec, link_index=None):
    if link_index is None:
        if isinstance(nelec, (int, numpy.number)):
            neleca = nelec//2
        else:
            neleca, nelecb = nelec
            assert (neleca == nelecb)
        link_index = cistring.gen_linkstr_index(range(norb), neleca)
    rdm1a = rdm.make_rdm1('FCItrans_rdm1a', cibra, ciket,
                          norb, nelec, link_index)
    rdm1b = rdm.make_rdm1('FCItrans_rdm1b', cibra, ciket,
                          norb, nelec, link_index)
    return rdm1a, rdm1b

@lib.with_doc(direct_spin1.trans_rdm1.__doc__)
def trans_rdm1(cibra, ciket, norb, nelec, link_index=None):
    rdm1a, rdm1b = trans_rdm1s(cibra, ciket, norb, nelec, link_index)
    return rdm1a + rdm1b

# dm[p,q,r,s] = <I|p^+ q r^+ s|J>
@lib.with_doc(direct_spin1.trans_rdm12.__doc__)
def trans_rdm12(cibra, ciket, norb, nelec, link_index=None, reorder=True):
    dm1, dm2 = rdm.make_rdm12('FCItdm12kern_sf', cibra, ciket,
                              norb, nelec, link_index, 2)
    if reorder:
        dm1, dm2 = rdm.reorder_rdm(dm1, dm2, True)
    return dm1, dm2

def energy(h1e, eri, fcivec, norb, nelec, link_index=None):
    h2e = direct_spin1.absorb_h1e(h1e, eri, norb, nelec, .5)
    ci1 = contract_2e(h2e, fcivec, norb, nelec, link_index)
    return numpy.dot(fcivec.ravel(), ci1.ravel())

def get_init_guess(norb, nelec, nroots, hdiag):
    if isinstance(nelec, (int, numpy.number)):
        nelecb = nelec//2
        neleca = nelec - nelecb
    else:
        neleca, nelecb = nelec
    na = cistring.num_strings(norb, neleca)
    nb = cistring.num_strings(norb, nelecb)

    init_strs = []
    iroot = 0
    for addr in numpy.argsort(hdiag):
        addra = addr // nb
        addrb = addr % nb
        if (addrb,addra) not in init_strs:  # avoid initial guess linear dependency
            init_strs.append((addra,addrb))
            iroot += 1
            if iroot >= nroots:
                break
    ci0 = []
    for addra,addrb in init_strs:
        x = numpy.zeros((na,nb))
        if addra == addrb:
            x[addra,addrb] = 1
        else:
            x[addra,addrb] = x[addrb,addra] = numpy.sqrt(.5)
        ci0.append(x.ravel().view(direct_spin1.FCIvector))

    # Add noise
    ci0[0][0 ] += 1e-5
    ci0[0][-1] -= 1e-5
    return ci0


###############################################################
# direct-CI driver
###############################################################

def kernel_ms0(fci, h1e, eri, norb, nelec, ci0=None, link_index=None,
               tol=None, lindep=None, max_cycle=None, max_space=None,
               nroots=None, davidson_only=None, pspace_size=None, hop=None,
               max_memory=None, verbose=None, ecore=0, **kwargs):
    if nroots is None: nroots = fci.nroots
    if davidson_only is None: davidson_only = fci.davidson_only
    if pspace_size is None: pspace_size = fci.pspace_size
    if max_memory is None:
        max_memory = fci.max_memory - lib.current_memory()[0]
    log = logger.new_logger(fci, verbose)

    nelec = direct_spin1._unpack_nelec(nelec, fci.spin)
    assert (0 <= nelec[0] <= norb and 0 <= nelec[1] <= norb)

    hdiag = fci.make_hdiag(h1e, eri, norb, nelec, compress=True)
    num_dets = hdiag.size
    pspace_size = min(num_dets, pspace_size)
    nroots = min(num_dets, nroots)
    na = cistring.num_strings(norb, nelec[0])
    addr = [0]
    pw = pv = None
    if pspace_size > 0 and norb < 64:
        addr, h0 = fci.pspace(h1e, eri, norb, nelec, hdiag, pspace_size)
        pw, pv = fci.eig(h0)

    if getattr(fci, 'sym_allowed_idx', None):
        # Remove symmetry forbidden elements
        sym_idx = numpy.hstack(fci.sym_allowed_idx)
        civec_size = sym_idx.size
    else:
        sym_idx = None
        civec_size = num_dets

    if max_memory < num_dets*6*8e-6:
        log.warn('Not enough memory for FCI solver. '
                 'The minimal requirement is %.0f MB', hdiag.size*60e-6)

    if pspace_size >= civec_size and ci0 is None and not davidson_only:
        e = []
        civec = []
        for i in range(pspace_size):
            c = numpy.empty(civec_size)
            c[addr] = pv[:,i]
            try:
                civec.append(_check_(c.reshape(na,na)))
            except ValueError:
                continue
            e.append(pw[i])
            if len(civec) >= nroots:
                break
        if nroots == 1:
            return e[0]+ecore, civec[0]
        else:
            return numpy.array(e)+ecore, civec
    pw = pv = h0 = None

    if sym_idx is None:
        precond = fci.make_precond(hdiag)
    else:
        precond = fci.make_precond(hdiag[sym_idx])

    h2e = fci.absorb_h1e(h1e, eri, norb, nelec, .5)
    if hop is None:
        cpu0 = [logger.process_clock(), logger.perf_counter()]
        def hop(c):
            hc = fci.contract_2e(h2e, c.reshape(na,na), norb, nelec, link_index)
            cpu0[:] = log.timer_debug1('contract_2e', *cpu0)
            return hc.ravel()

    def init_guess():
        if callable(getattr(fci, 'get_init_guess', None)):
            return fci.get_init_guess(norb, nelec, nroots, hdiag)
        else:
            x0 = []
            for i in range(nroots):
                x = numpy.zeros((na,na))
                addra = addr[i] // na
                addrb = addr[i] % na
                if addra == addrb:
                    x[addra,addrb] = 1
                else:
                    x[addra,addrb] = x[addrb,addra] = numpy.sqrt(.5)
                x0.append(x.ravel())
            return x0

    if ci0 is None:
        ci0 = init_guess  # lazy initialization to reduce memory footprint
    elif not callable(ci0):
        if isinstance(ci0, numpy.ndarray):
            ci0 = [ci0.ravel()]
        else:
            ci0 = [x.ravel() for x in ci0]
        if sym_idx is not None and ci0[0].size != civec_size:
            ci0 = [x[sym_idx] for x in ci0]
        if len(ci0) < nroots:
            ci0.extend(init_guess()[len(ci0):])

    if tol is None: tol = fci.conv_tol
    if lindep is None: lindep = fci.lindep
    if max_cycle is None: max_cycle = fci.max_cycle
    if max_space is None: max_space = fci.max_space
    tol_residual = getattr(fci, 'conv_tol_residual', None)

    with lib.with_omp_threads(fci.threads):
        e, c = fci.eig(hop, ci0, precond, tol=tol, lindep=lindep,
                       max_cycle=max_cycle, max_space=max_space, nroots=nroots,
                       max_memory=max_memory, verbose=log, follow_state=True,
                       tol_residual=tol_residual, **kwargs)
    return e+ecore, c

def _check_(c):
    c = lib.transpose_sum(c, inplace=True)
    c *= .5
    norm = numpy.linalg.norm(c)
    if abs(norm-1) > 1e-6:
        raise ValueError('State not singlet %g' % (norm - 1))
    return c/norm


class FCISolver(direct_spin1.FCISolver):

    make_hdiag = staticmethod(make_hdiag)

    def contract_1e(self, f1e, fcivec, norb, nelec, link_index=None, **kwargs):
        return contract_1e(f1e, fcivec, norb, nelec, link_index, **kwargs)

    def contract_2e(self, eri, fcivec, norb, nelec, link_index=None, **kwargs):
        return contract_2e(eri, fcivec, norb, nelec, link_index, **kwargs)

    def get_init_guess(self, norb, nelec, nroots, hdiag):
        return get_init_guess(norb, nelec, nroots, hdiag)

    def kernel(self, h1e, eri, norb, nelec, ci0=None,
               tol=None, lindep=None, max_cycle=None, max_space=None,
               nroots=None, davidson_only=None, pspace_size=None,
               orbsym=None, wfnsym=None, ecore=0, **kwargs):
        if nroots is None: nroots = self.nroots
        if self.verbose >= logger.WARN:
            self.check_sanity()
        assert self.spin is None or self.spin == 0
        self.norb = norb
        self.nelec = nelec
        link_index = direct_spin1._unpack(norb, nelec, None)
        e, c = kernel_ms0(self, h1e, eri, norb, nelec, ci0, link_index,
                          tol, lindep, max_cycle, max_space, nroots,
                          davidson_only, pspace_size, ecore=ecore, **kwargs)
        self.eci = e

        na = link_index[0].shape[0]
        if nroots > 1:
            self.ci = [
                _check_(x.reshape(na,na)).view(direct_spin1.FCIvector) for x in c]
        else:
            self.ci = _check_(c.reshape(na,na)).view(direct_spin1.FCIvector)
        return self.eci, self.ci

    def energy(self, h1e, eri, fcivec, norb, nelec, link_index=None):
        h2e = self.absorb_h1e(h1e, eri, norb, nelec, .5)
        ci1 = self.contract_2e(h2e, fcivec, norb, nelec, link_index)
        return numpy.dot(fcivec.reshape(-1), ci1.reshape(-1))

    def make_rdm1s(self, fcivec, norb, nelec, link_index=None):
        return make_rdm1s(fcivec, norb, nelec, link_index)

    def make_rdm1(self, fcivec, norb, nelec, link_index=None):
        return make_rdm1(fcivec, norb, nelec, link_index)

    @lib.with_doc(make_rdm12.__doc__)
    def make_rdm12(self, fcivec, norb, nelec, link_index=None, reorder=True):
        return make_rdm12(fcivec, norb, nelec, link_index, reorder)

    def trans_rdm1s(self, cibra, ciket, norb, nelec, link_index=None):
        return trans_rdm1s(cibra, ciket, norb, nelec, link_index)

    def trans_rdm1(self, cibra, ciket, norb, nelec, link_index=None):
        return trans_rdm1(cibra, ciket, norb, nelec, link_index)

    @lib.with_doc(trans_rdm12.__doc__)
    def trans_rdm12(self, cibra, ciket, norb, nelec, link_index=None,
                    reorder=True):
        return trans_rdm12(cibra, ciket, norb, nelec, link_index, reorder)

    def gen_linkstr(self, norb, nelec, tril=True, spin=None):
        if isinstance(nelec, (int, numpy.number)):
            neleca = nelec//2
        else:
            neleca, nelecb = nelec
            assert (neleca == nelecb)
        if tril:
            link_index = cistring.gen_linkstr_index_trilidx(range(norb), neleca)
        else:
            link_index = cistring.gen_linkstr_index(range(norb), neleca)
        return link_index

FCI = FCISolver


if __name__ == '__main__':
    from functools import reduce
    from pyscf import gto
    from pyscf import scf

    mol = gto.Mole()
    mol.verbose = 0
    mol.output = None#"out_h2o"
    mol.atom = [
        ['H', ( 1.,-1.    , 0.   )],
        ['H', ( 0.,-1.    ,-1.   )],
        ['H', ( 1.,-0.5   ,-1.   )],
        ['H', ( 0.,-0.5   ,-1.   )],
        ['H', ( 0.,-0.5   ,-0.   )],
        ['H', ( 0.,-0.    ,-1.   )],
        ['H', ( 1.,-0.5   , 0.   )],
        ['H', ( 0., 1.    , 1.   )],
    ]

    mol.basis = {'H': 'sto-3g'}
    mol.build()

    m = scf.RHF(mol)
    ehf = m.scf()

    cis = FCISolver(mol)
    norb = m.mo_coeff.shape[1]
    nelec = mol.nelectron
    h1e = reduce(numpy.dot, (m.mo_coeff.T, m.get_hcore(), m.mo_coeff))
    eri = ao2mo.incore.general(m._eri, (m.mo_coeff,)*4, compact=False)
    e, c = cis.kernel(h1e, eri, norb, nelec)
    print(e - -15.9977886375)
    print('t',logger.process_clock())
